from ftw.solr.connection import SolrResponse
from ftw.solr.tests.utils import get_data
import socket
import unittest


class TestResponse(unittest.TestCase):

    def test_response_from_exception(self):
        exception = socket.error(61, 'Connection refused')
        resp = SolrResponse(exception=exception)
        self.assertFalse(resp.is_ok())
        self.assertEqual(resp.exception, exception)

    def test_response_from_bad_body(self):
        resp = SolrResponse(body='bad', status=200)
        self.assertFalse(resp.is_ok())
        self.assertEqual(resp.status, -1)

    def test_response_from_bad_request(self):
        body = get_data('bad_request.json')
        resp = SolrResponse(body=body, status=400)
        self.assertFalse(resp.is_ok())
        self.assertEqual(resp.http_status, 400)
        self.assertEqual(resp.status, 400)
        self.assertEqual(resp[u'error'][u'code'], 400)

    def test_response_from_search_request(self):
        body = get_data('search.json')
        resp = SolrResponse(body=body, status=200)
        self.assertTrue(resp.is_ok())
        self.assertEqual(len(resp.docs), 3)
        self.assertEqual(resp.num_found, 3)
        self.assertEqual(resp.start, 0)
        self.assertEqual(len(resp.facets), 0)
        self.assertEqual(resp.facets, {})

    def test_response_from_search_request_with_facets(self):
        body = get_data('search_with_facets.json')
        resp = SolrResponse(body=body, status=200)
        self.assertTrue(resp.is_ok())
        self.assertEqual(len(resp.docs), 3)
        self.assertEqual(resp.num_found, 3)
        self.assertEqual(resp.start, 0)
        self.assertEqual(len(resp.facets), 2)
        self.assertEqual(resp.facets,
                         {u'portal_type': {u'Folder': 2, u'Document': 1},
                          u'review_state': {u'private': 3}})

    def test_response_from_search_with_highlighting(self):
        body = get_data('highlighting.json')
        resp = SolrResponse(body=body, status=200)
        self.assertTrue(resp.is_ok())
        self.assertIn('highlighting', resp)

    def test_response_from_schema_request(self):
        body = get_data('schema.json')
        resp = SolrResponse(body=body, status=200)
        self.assertTrue(resp.is_ok())
        self.assertIn('schema', resp)



# -*- coding: utf-8 -*-
from ftw.solr.interfaces import ISolrSearch
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING
from mock import MagicMock
from plone import api
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from zope.component import getUtility
import unittest


class TestSearch(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.solr = getUtility(ISolrSearch)
        self.solr._manager = MagicMock(name='SolrConnectionManager')
        self.conn = MagicMock(name='SolrConnection')
        self.solr.manager.connection = self.conn

    def test_security_filter_for_anonymous(self):
        logout()
        self.assertEqual(
            self.solr.security_filter(), u'allowedRolesAndUsers:Anonymous')

    def test_security_filter_for_authenticated_user(self):
        self.assertEqual(
            self.solr.security_filter(),
            u'allowedRolesAndUsers:(Member OR Authenticated OR Anonymous OR '
            u'user\:AuthenticatedUsers OR user\:test_user_1_)')

    def test_security_filter_contains_roles(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Member', 'Reader', 'Editor'])
        security_filter = self.solr.security_filter()
        self.assertIn(u'Member', security_filter)
        self.assertIn(u'Reader', security_filter)
        self.assertIn(u'Editor', security_filter)

    def test_security_filter_contains_groups(self):
        api.group.create(groupname='staff')
        api.group.add_user(groupname='staff', username=TEST_USER_ID)
        # Update groups in user object
        self.layer['portal'].acl_users._getGroupsForPrincipal(
            api.user.get_current().getUser())
        self.assertIn(u'user\\:staff', self.solr.security_filter())

    def test_security_filter_contains_current_user(self):
        self.assertIn(
            u'user\\:%s' % TEST_USER_ID,
            self.solr.security_filter())

    def test_security_filter_quotes_roles_and_users_with_spaces(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Records Manager', 'Member'])
        security_filter = self.solr.security_filter()
        self.assertIn('"Records Manager"', security_filter)

    def test_search_with_query(self):
        self.solr.search(query=u'Title:foo')
        args, kwargs = self.conn.search.call_args
        self.assertEqual(args[0][u'query'], u'Title:foo')

    def test_search_with_single_filter(self):
        self.solr.search(filters=u'portal_type:File')
        args, kwargs = self.conn.search.call_args
        self.assertEqual(
            args[0][u'filter'],
            [self.solr.security_filter(), u'portal_type:File'])

    def test_search_with_multiple_filters(self):
        self.solr.search(
            filters=[u'portal_type:File', u'review_state:published'])
        args, kwargs = self.conn.search.call_args
        self.assertEqual(
            args[0][u'filter'],
            [
                self.solr.security_filter(),
                u'portal_type:File',
                u'review_state:published',
            ])

    def test_search_with_pagination(self):
        self.solr.search(start=20, rows=10)
        args, kwargs = self.conn.search.call_args
        self.assertEqual(args[0][u'offset'], 20)
        self.assertEqual(args[0][u'limit'], 10)

    def test_search_with_sort(self):
        self.solr.search(sort='modified desc')
        args, kwargs = self.conn.search.call_args
        self.assertEqual(args[0][u'sort'], 'modified desc')

    def test_search_with_custom_parameters(self):
        self.solr.search(**{u'hl': u'on', u'hl.fl': u'SearchableText'})
        args, kwargs = self.conn.search.call_args
        self.assertEqual(
            args[0][u'params'],
            {u'hl': u'on', u'hl.fl': u'SearchableText'},
            )

    def test_search_with_custom_request_handler(self):
        self.solr.search(request_handler='/query')
        args, kwargs = self.conn.search.call_args
        self.assertEqual(kwargs, {'request_handler': u'/query'})

    def test_search_contains_security_filters(self):
        self.solr.search()
        args, kwargs = self.conn.search.call_args
        self.assertEqual(
            args[0][u'filter'],
            [u'allowedRolesAndUsers:(Member OR Authenticated OR Anonymous OR '
             u'user\\:AuthenticatedUsers OR user\\:test_user_1_)'])

    def test_search_does_not_allow_unrestricted_keyword(self):
        with self.assertRaises(TypeError):
            self.solr.search(unrestricted=True)

    def test_unrestricted_search_does_not_contain_security_filters(self):
        self.solr.unrestricted_search()
        args, kwargs = self.conn.search.call_args
        self.assertEqual(args[0][u'filter'], [])



import os.path


def get_data(filename):
    """Return content from a file in the test data folder """
    filename = os.path.join(os.path.dirname(__file__), 'data', filename)
    return open(filename, 'r').read()


def normalize_whitespaces(text):
    return ' '.join(text.strip().split())


class MockHTTPResponse(object):
    def __init__(self, status=200, body=''):
        self.status = status
        self.body = body

    def read(self):
        return self.body


class MockBlob(object):
    def __init__(self, path='/folder/file'):
        self.path = path

    def committed(self):
        return self.path



from ftw.solr.connection import local_data
from ftw.solr.connection import SolrConnection
from ftw.solr.connection import SolrConnectionConfig
from ftw.solr.connection import SolrConnectionManager
from ftw.solr.connection import SolrResponse
from ftw.solr.tests.utils import get_data
from ftw.solr.tests.utils import MockHTTPResponse
from ftw.solr.tests.utils import MockBlob
from ftw.solr.interfaces import ISolrConnectionConfig
from mock import patch
from mock import MagicMock
from zope.component import provideUtility
from plone.testing import zca
import unittest
import transaction
import socket


class TestConnection(unittest.TestCase):

    def test_connection_initialization(self):
        conn = SolrConnection(host='mysolrserver', base='/solr/mycore')
        self.assertEqual(conn.conn.host, 'mysolrserver')
        self.assertEqual(conn.conn.port, 8983)
        self.assertEqual(conn.base, '/solr/mycore')

    @patch('ftw.solr.connection.HTTPConnection', autospec=True)
    def test_successful_request_returns_response(self, MockHTTPConnection):
        MockHTTPConnection.return_value.request.return_value = None
        MockHTTPConnection.return_value.getresponse.return_value = (
            MockHTTPResponse(body=get_data('search.json')))
        conn = SolrConnection(base='/solr/mycore')
        resp = conn.request('POST', '/select', data='{"query":"*:*" }')
        self.assertEqual(resp.http_status, 200)

    @patch('ftw.solr.connection.HTTPConnection', autospec=True)
    def test_broken_connection_returns_response(self, MockHTTPConnection):
        exception = socket.error(61, 'Connection refused')
        MockHTTPConnection.return_value.request.side_effect = exception
        conn = SolrConnection(base='/solr/mycore')
        resp = conn.request('POST', '/select', data='{"query":"*:*" }')
        self.assertEqual(resp.exception, exception)

    @patch('ftw.solr.connection.HTTPConnection', autospec=True)
    def test_broken_connection_returns_response_after_reconnect(
            self, MockHTTPConnection):
        exception = socket.error(61, 'Connection refused')
        MockHTTPConnection.return_value.request.side_effect = [exception, None]
        MockHTTPConnection.return_value.getresponse.return_value = (
            MockHTTPResponse(body=get_data('search.json')))
        conn = SolrConnection(base='/solr/mycore')
        resp = conn.request('POST', '/select', data='{"query":"*:*" }')
        self.assertEqual(resp.http_status, 200)
        self.assertEqual(resp.exception, None)

    def test_add_operation_queues_update_command(self):
        conn = SolrConnection()
        conn.add({'id': '1'})
        self.assertEqual(conn.update_commands, ['"add": {"doc": {"id": "1"}}'])

    def test_extract_operation_queues_extract_command(self):
        conn = SolrConnection()
        conn.extract('Blob', 'SearchableText', {'id': '1'},
                     'application/octet-stream')
        self.assertEqual(
            conn.extract_commands,
            [('Blob', 'SearchableText', {'id': '1'}, 'application/octet-stream')])

    def test_delete_operation_queues_update_command(self):
        conn = SolrConnection()
        conn.delete('1')
        self.assertEqual(conn.update_commands, ['"delete": {"id": "1"}'])

    def test_delete_by_query_operation_queues_update_command(self):
        conn = SolrConnection()
        conn.delete_by_query('*:*')
        self.assertEqual(conn.update_commands, ['"delete": {"query": "*:*"}'])

    def test_commit_operation_queues_update_command_and_flushes_queue(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.flush = MagicMock(name='flush')
        conn.commit()
        conn.flush.assert_called_once_with(extract_after_commit=True)
        self.assertEqual(
            conn.update_commands,
            ['"commit": {"softCommit": true, "waitSearcher": true}'])

    def test_optimize_operation_queues_update_command_and_flushes_queue(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.flush = MagicMock(name='flush')
        conn.optimize()
        conn.flush.assert_called_once()
        self.assertEqual(
            conn.update_commands, ['"optimize": {"waitSearcher": true}'])

    def test_flush_operation_posts_update_commands_and_clears_queue(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.post = MagicMock(name='post', return_value=SolrResponse(
            body='{"responseHeader":{"status":0}}', status=200))
        conn.add({'id': '1'})
        conn.flush()
        conn.post.assert_called_once_with(
            '/update', data='{"add": {"doc": {"id": "1"}}}', log_error=False)
        self.assertEqual(conn.update_commands, [])

    def test_flush_operation_posts_extract_commands_and_clears_queue(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.post = MagicMock(name='post')
        conn.post.return_value.body.get.return_value = 'The searchable text'
        tr = transaction.begin()
        conn.extract(MockBlob(), 'SearchableText', {'id': '1'},
                     'application/octet-stream')
        conn.flush()
        tr.commit()
        args, kwargs = conn.post.call_args_list[0]
        self.assertEqual(
            args,
            ('/update/extract?extractOnly=true&stream.file=%2Ffolder%2Ffile',))
        self.assertEqual(
            kwargs,
            {'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
             'log_error': False},
        )
        args, kwargs = conn.post.call_args_list[1]
        self.assertEqual(args, ('/update',))
        self.assertEqual(
            kwargs,
            {'data': '{"add": {"doc": {"id": "1", "SearchableText": {"set": "T'
                     'he searchable text"}}}}',
             'log_error': False})
        self.assertEqual(conn.extract_commands, [])

    def test_flush_operation_posts_filtered_extract_commands(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.post = MagicMock(name='post')
        conn.post.return_value.body.get.return_value = 'The searchable text'
        tr = transaction.begin()
        conn.extract(MockBlob('file1'), 'SearchableText', {'UID': '1'},
                     'application/octet-stream')
        conn.extract(MockBlob('file2'), 'SearchableText', {'UID': '2'},
                     'application/octet-stream')
        conn.extract(MockBlob('file3'), 'SearchableText', {'UID': '2'},
                     'application/octet-stream')
        conn.extract(MockBlob('file4'), 'OtherField', {'UID': '1'},
                     'application/octet-stream')
        conn.flush()
        tr.commit()

        self.assertEqual(6, len(conn.post.call_args_list))

        self.assertEqual(
            (('/update/extract?extractOnly=true&stream.file=file1',),
             {'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
              'log_error': False}),
            conn.post.call_args_list[0])

        self.assertEqual(
            (('/update',),
             {'data': '{"add": {"doc": {"UID": "1", "SearchableText": {"set": '
                      '"The searchable text"}}}}',
              'log_error': False}),
            conn.post.call_args_list[1])

        self.assertEqual(
            (('/update/extract?extractOnly=true&stream.file=file3',),
             {'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
              'log_error': False}),
            conn.post.call_args_list[2])

        self.assertEqual(
            (('/update',),
             {'data': '{"add": {"doc": {"UID": "2", "SearchableText": {"set": '
                      '"The searchable text"}}}}',
              'log_error': False}),
            conn.post.call_args_list[3])

        self.assertEqual(
            (('/update/extract?extractOnly=true&stream.file=file4',),
             {'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
              'log_error': False}),
            conn.post.call_args_list[4])

        self.assertEqual(
            (('/update',),
             {'data': '{"add": {"doc": {"OtherField": {"set": '
                      '"The searchable text"}, "UID": "1"}}}',
              'log_error': False}),
            conn.post.call_args_list[5])

        self.assertEqual(conn.extract_commands, [])

    def test_flush_operation_posts_extract_commands_with_blobs_if_configured(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.upload_blobs = True
        conn.post = MagicMock(name='post')
        conn.post_chunked = MagicMock(name='post_chunked')
        conn.post_chunked.return_value.body.get.return_value = 'The searchable text'
        tr = transaction.begin()
        blob = MockBlob()
        conn.extract(blob, 'SearchableText', {'id': '1'},
                     'application/octet-stream')
        conn.flush()
        tr.commit()
        args, kwargs = conn.post_chunked.call_args_list[0]
        self.assertEqual(
            args,
            ('/update/extract?extractOnly=true', blob))
        self.assertEqual(
            kwargs,
            {'content_type': 'application/octet-stream',
             'log_error': False},
        )
        args, kwargs = conn.post.call_args_list[0]
        self.assertEqual(args, ('/update',))
        self.assertEqual(
            kwargs,
            {'data': '{"add": {"doc": {"id": "1", "SearchableText": {"set": "T'
                     'he searchable text"}}}}',
             'log_error': False})
        self.assertEqual(conn.extract_commands, [])

    def test_flush_operation_without_after_commit_hook(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.post = MagicMock(name='post')
        conn.post.return_value.body.get.return_value = 'The searchable text'
        conn.extract(MockBlob(), 'SearchableText', {'id': '1'},
                     'application/octet-stream')
        conn.flush(extract_after_commit=False)

        args, kwargs = conn.post.call_args_list[0]
        self.assertEqual(
            args,
            ('/update/extract?extractOnly=true&stream.file=%2Ffolder%2Ffile',))
        self.assertEqual(
            kwargs,
            {'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
             'log_error': False},
        )
        args, kwargs = conn.post.call_args_list[1]
        self.assertEqual(args, ('/update',))
        self.assertEqual(
            kwargs,
            {'data': '{"add": {"doc": {"id": "1", "SearchableText": {"set": "T'
                     'he searchable text"}}}}',
             'log_error': False})

        self.assertEqual(conn.extract_commands, [])

    def test_abort_operation_clears_queue(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.add({'id': '1'})
        conn.extract(MockBlob(), 'SearchableText', {'id': '1'},
                     'application/octet-stream')
        conn.abort()
        self.assertEqual(conn.update_commands, [])
        self.assertEqual(conn.extract_commands, [])

    def test_search_operation_posts_search_request(self):
        conn = SolrConnection(base='/solr/mycore')
        conn.post = MagicMock(name='post')
        conn.search({'query': '*:*'})
        conn.post.assert_called_once_with('/select', '{"query": "*:*"}')


class TestConnectionManager(unittest.TestCase):

    layer = zca.UNIT_TESTING

    def setUp(self):
        if hasattr(local_data, 'connection'):
            del local_data.connection

    def test_manager_creates_new_connection(self):
        manager = SolrConnectionManager()
        config = SolrConnectionConfig('myhost', 8983, '/solr/mycore')
        provideUtility(config, ISolrConnectionConfig)
        conn = manager.connection
        self.assertEqual(conn.host, 'myhost')
        self.assertEqual(conn.port, 8983)
        self.assertEqual(conn.base, '/solr/mycore')

    def test_manager_reuses_existing_connection(self):
        manager = SolrConnectionManager()
        config = SolrConnectionConfig('myhost', 8983, '/solr/mycore')
        provideUtility(config, ISolrConnectionConfig)
        conn1 = manager.connection
        conn2 = manager.connection
        self.assertEqual(conn1, conn2)

    def test_manager_returns_none_with_no_config(self):
        manager = SolrConnectionManager()
        conn = manager.connection
        self.assertEqual(conn, None)

    def test_manager_creates_new_schema(self):
        manager = SolrConnectionManager()
        config = SolrConnectionConfig('myhost', 8983, '/solr/mycore')
        provideUtility(config, ISolrConnectionConfig)
        schema = manager.schema
        self.assertIn('unique_key', dir(schema))
        self.assertIn('fields', dir(schema))

    def test_manager_reuses_existing_schema(self):
        manager = SolrConnectionManager()
        config = SolrConnectionConfig('myhost', 8983, '/solr/mycore')
        provideUtility(config, ISolrConnectionConfig)
        schema1 = manager.schema
        schema1.unique_key = u'UID'  # Make it a 'valid' schema
        schema2 = manager.schema
        self.assertEqual(schema1, schema2)



# -*- coding: utf-8 -*-
from ftw.solr.interfaces import PLONE51
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2
import ftw.solr
import plone.app.contenttypes


class FtwSolrLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        if not PLONE51:
            import collective.indexing
            self.loadZCML(package=collective.indexing)
            z2.installProduct(app, 'collective.indexing')

        z2.installProduct(app, 'Products.DateRecurringIndex')
        self.loadZCML(package=plone.app.contenttypes)
        self.loadZCML(package=ftw.solr)
        z2.installProduct(app, 'ftw.solr')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.contenttypes:default')
        applyProfile(portal, 'ftw.solr:default')


class FtwSolrATLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        if not PLONE51:
            import collective.indexing
            self.loadZCML(package=collective.indexing)
            z2.installProduct(app, 'collective.indexing')

        self.loadZCML(package=ftw.solr)
        z2.installProduct(app, 'ftw.solr')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.solr:default')


FTW_SOLR_FIXTURE = FtwSolrLayer()
FTW_SOLR_AT_FIXTURE = FtwSolrATLayer()

FTW_SOLR_INTEGRATION_TESTING = IntegrationTesting(
    bases=(FTW_SOLR_FIXTURE,),
    name='FtwSolrLayer:IntegrationTesting'
)

FTW_SOLR_AT_INTEGRATION_TESTING = IntegrationTesting(
    bases=(FTW_SOLR_AT_FIXTURE,),
    name='FtwSolrATLayer:IntegrationTesting'
)

FTW_SOLR_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FTW_SOLR_FIXTURE,),
    name='FtwSolrLayer:FunctionalTesting'
)



from ftw.solr.interfaces import PLONE51
from plone.indexer.interfaces import IIndexer
from Products.Archetypes.CatalogMultiplex import CatalogMultiplex
from Products.Archetypes.config import TOOL_NAME
from Products.Archetypes.utils import isFactoryContained
from Products.CMFCore.CMFCatalogAware import CMFCatalogAware
from Products.CMFCore.interfaces import ICatalogTool
from Products.CMFCore.interfaces._content import ICatalogAware
from Products.CMFCore.utils import getToolByName
from zope.component import queryMultiAdapter

if not PLONE51:
    # Make sure collective.indexing patches are applied before our patches
    import collective.indexing.monkey  # noqa


def is_index_value_equal(old, new):
    """Compares an old and a new index value.
    If the value is of type list, it is compared as set since this usually is
    used in a KeywordIndex, where the order is not relevant.
    """
    if type(old) != type(new):
        return False
    if isinstance(old, list):
        return set(old) == set(new)
    return old == new


def is_index_up_to_date(catalog, obj, index_name):
    """Checks whether the passed index (`index_name`) of the passed object
    (`obj`) is up to date in the passed catalog (`catalog`).
    """
    indexer = queryMultiAdapter((obj, catalog), IIndexer, name=index_name)
    if indexer is None:
        indexer = getattr(obj, index_name, None)

    if indexer is None:
        # We cannot re-generate the index data, therfore we
        # act as if it is outdated
        return False

    else:
        path = '/'.join(obj.getPhysicalPath())
        rid = catalog.getrid(path)
        if rid is None:
            # the object was not yet indexed
            return False

        indexed_values = catalog.getIndexDataForRID(catalog.getrid(path))
        value_before = indexed_values.get(index_name, None)
        value_after = indexer()

        if any(r in value_after for r in ('Anonymous', 'Authenticated')):
            # Account for the shortcut in the allowedRolesAndUsers indexer
            # regarding these system roles:
            #
            # If any of these roles are present in the allowed roles, the
            # resulting value to be indexed gets simplified to just that
            # role name.
            #
            # This can lead to situations where the effective set of allowed
            # roles and principals changed, but the indexed value stays the
            # same. That's fine for that specific object, but the change may
            # make it necessary to reindex security for a child further down
            # the line.
            #
            # But because we use is_index_up_to_date() to determine
            # when to stop recursion, we would stop recursion too early in
            # that case.
            #
            # We therefore always assume the indexed value has changed if we
            # see any of these two roles in the new indexed value.
            return False

        if not is_index_value_equal(value_before, value_after):
            return False

    return True


def recursive_index_security(catalog, obj, skip_self=False):
    """This function recursively reindexes the security indexes for an object
    in a specific catalog.

    It does so by walking down the tree and checking for every object whether
    its security indexes are already up to date. If they are up to date, it
    stops recursion.

    The aim of terminating recursion early is to drastically improve
    performance for large trees. This also avoids unnecessarily reindexing the
    same objects multiple times if this function is called for several
    overlapping subtrees.

    Reasoning why terminating recursion early is safe:
    --------------------------------------------------

    CMFCatalogAware.reindexObjectSecurity() needs to be recursive because
    changes to an object's security may affect contained subobjects.

    Indexed security for objects in Plone can only be influenced by their
    parents via some kind of inheritance. There's exactly two inheritance
    mechanisms in play:

    - Acquisition of permissions via the obj's security settings (manage_access)
    - Inheritance of local roles

    (An obj's security settings can indirectly be managed via workflows. This
    doesn't matter here though, it's irrelevant how exactly they came to be).

    In both of these cases, only the immediate parent of a subobject is
    relevant for inheritance. Therefore, if an objects indexed security didn't
    experience any changes, neither can any of its subobjects - recursively.

    Because of this, downstream propagation can be stopped as soon as an
    object is encountered whose indexed security didn't change.

    (Note: This assumes that the indexed allowed roles of an object correspond
    exactly to the real allowed roles - if the real roles change, so does the
    indexed value. This isn't strictly the case - we take care of that in
    is_index_up_to_date(), see the corresponding comment there.)

    ---

    In the case of workflow changes, reindexObjectSecurity() still needs to
    be called for every object that is *directly* affected by the change:

    - switched to a different WF
    - to a different WF state
    - changes in the WF state's security settings
    - object moved to/out of a placeful WF).

    Identifying these objects is usually easy, and you can't and must not rely
    on reinexObjectSecurity's recursion to pick them up. Instead, they can be
    determined by querying for the relevant criteria, e.g. objects that have
    a particular WF.

    Recursion will then take care of updating security indexes for affected
    subobjects that *don't* meet the criteria for being directly affected (
    like subobjects with a different workflow).
    """
    indexes_to_update = []

    # Since objectValues returns all objects, including placefulworkflow policy
    # objects, we have to check if the object is Catalog aware.
    if not ICatalogAware.providedBy(obj):
        return

    for index_name in obj._cmf_security_indexes:
        if not is_index_up_to_date(catalog, obj, index_name):
            indexes_to_update.append(index_name)

    if indexes_to_update:
        if not skip_self:
            obj.reindexObject(idxs=indexes_to_update)

        # We assume that if the parent is up to date, all children are too.
        # Therefore we stop recursion as soon as we encounter an object
        # whose security indexes are up to date.
        for subobj in obj.objectValues():
            recursive_index_security(catalog, subobj)


def ftw_solr_CatalogMultiplex_reindexObjectSecurity(self, skip_self=False):
    """Update security information in all registered catalogs.
    """
    if isFactoryContained(self):
        return
    at = getToolByName(self, TOOL_NAME, None)
    if at is None:
        return

    catalogs = [c for c in at.getCatalogsByType(self.meta_type)
                if ICatalogTool.providedBy(c)]

    for catalog in catalogs:
        recursive_index_security(catalog, self, skip_self=skip_self)


def ftw_solr_CatalogAware_reindexObjectSecurity(self, skip_self=False):
    """Reindex security-related indexes on the object.
    """
    catalog = self._getCatalogTool()
    if catalog is None:
        return

    s = getattr(self, '_p_changed', 0)
    recursive_index_security(catalog, self, skip_self=skip_self)
    if s is None:
        self._p_deactivate()


CMFCatalogAware.reindexObjectSecurity = ftw_solr_CatalogAware_reindexObjectSecurity  # noqa
CatalogMultiplex.reindexObjectSecurity = ftw_solr_CatalogMultiplex_reindexObjectSecurity  # noqa



from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrIndexHandler
from ftw.solr.interfaces import ISolrIndexQueueProcessor
from ftw.solr.interfaces import ISolrSettings
from logging import getLogger
from plone.registry.interfaces import IRegistry
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer


logger = getLogger('ftw.solr.indexer')


@implementer(ISolrIndexQueueProcessor)
class SolrIndexQueueProcessor(object):
    """A queue processor for solr """

    _manager = None

    def is_enabled(self):
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(ISolrSettings)
        return settings.enabled

    def index(self, obj, attributes=None):
        """Index the given object."""
        if self.is_enabled():
            handler = getMultiAdapter((obj, self.manager), ISolrIndexHandler)
            handler.add(attributes)

    def reindex(self, obj, attributes=None, update_metadata=1):
        """Reindex the given object."""
        self.index(obj, attributes)

    def unindex(self, obj):
        """Unindex the given object."""
        if self.is_enabled():
            handler = getMultiAdapter((obj, self.manager), ISolrIndexHandler)
            handler.delete()

    def begin(self):
        """Called before processing of the queue is started."""
        pass

    def commit(self):
        """Called after processing of the queue has ended."""
        conn = self.manager.connection
        if conn is None:
            return
        if self.is_enabled():
            conn.commit()

    def abort(self):
        """Called if processing of the queue needs to be aborted."""
        conn = self.manager.connection
        if conn is None:
            return
        if self.is_enabled():
            conn.abort()

    @property
    def manager(self):
        if self._manager is None:
            self._manager = queryUtility(ISolrConnectionManager)
        return self._manager



from ftw.solr.connection import SolrResponse
from ftw.solr.schema import SolrSchema
from ftw.solr.tests.utils import get_data
from mock import MagicMock
from mock import PropertyMock
import unittest


class TestSchema(unittest.TestCase):

    def setUp(self):
        conn = MagicMock(name='SolrConnection')
        conn.get = MagicMock(name='get', return_value=SolrResponse(
            body=get_data('schema.json'), status=200))
        manager = MagicMock(name='SolrConnectionManager')
        type(manager).connection = PropertyMock(return_value=conn)
        self.schema = SolrSchema(manager)

    def test_schema_retrieval(self):
        schema = self.schema
        self.assertEqual(schema.version, 1.6)
        self.assertEqual(schema.unique_key, 'UID')
        self.assertItemsEqual(
            schema.fields.keys(),
            [
                u'Title',
                u'modified',
                u'SearchableText',
                u'allowedRolesAndUsers',
                u'_version_',
                u'_root_',
                u'UID',
                u'path',
                u'path_depth',
            ])
        self.assertEqual(schema.copy_fields.keys(), [])
        self.assertEqual(schema.dynamic_fields.keys(), [])
        self.assertItemsEqual(
            schema.field_types.keys(),
            [
                u'boolean',
                u'string',
                u'plong',
                u'pint',
                u'text',
                u'pdate',
            ])

    def test_get_field_class_of_existing_field(self):
        self.assertEqual(self.schema.field_class('UID'), 'solr.StrField')

    def test_get_field_class_of_not_existing_field(self):
        self.assertEqual(self.schema.field_class('foo'), None)



# -*- coding: utf-8 -*-
from ftw.solr.indexer import SolrIndexQueueProcessor
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrIndexHandler
from ftw.solr.interfaces import ISolrSettings
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING
from mock import MagicMock
from mock import PropertyMock
from plone.app.testing import popGlobalRegistry
from plone.app.testing import pushGlobalRegistry
from plone.registry.interfaces import IRegistry
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.component import queryUtility
from zope.interface import Interface
import unittest


class MockSolrIndexHandler(object):

    add = MagicMock(name='add')
    delete = MagicMock(name='delete')

    def __init__(self, context, manager):
        self.manger = manager


class TestSolrIndexQueueProcessor(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.indexer = SolrIndexQueueProcessor()
        self.indexer._manager = MagicMock(name='SolrConnectionManager')
        pushGlobalRegistry(self.layer['portal'])
        provideAdapter(
            MockSolrIndexHandler,
            adapts=(Interface, MagicMock),
            provides=ISolrIndexHandler,
        )
        conn = MagicMock(name='SolrConnection')
        conn.commit = MagicMock(name='commit')
        type(self.indexer._manager).connection = PropertyMock(
            return_value=conn)

    def tearDown(self):
        popGlobalRegistry(self.layer['portal'])

    def disable_indexing(self):
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(ISolrSettings)
        settings.enabled = False

    def test_index_object_without_providing_attributes_calls_add(self):
        self.indexer.index(object())
        MockSolrIndexHandler.add.assert_called_with(None)

    def test_index_object_with_provided_attributes_calls_add(self):
        self.indexer.index(object(), attributes=['path'])
        MockSolrIndexHandler.add.assert_called_with(['path'])

    def test_reindex_object_without_providing_attributes_calls_add(self):
        self.indexer.reindex(object())
        MockSolrIndexHandler.add.assert_called_with(None)

    def test_reindex_object_with_provided_attributes_calls_add(self):
        self.indexer.reindex(object(), attributes=['path', 'Title'])
        MockSolrIndexHandler.add.assert_called_with(['path', 'Title'])

    def test_unindex_object_calls_delete(self):
        self.indexer.unindex(object())
        MockSolrIndexHandler.delete.assert_called_with()

    def test_begin_does_nothing(self):
        self.indexer.manager.connection.commit.assert_not_called()
        self.indexer.manager.connection.abort.assert_not_called()

    def test_commit_calls_connection_commit(self):
        self.indexer.commit()
        self.indexer.manager.connection.commit.assert_called_with()

    def test_abort_calls_connection_abort(self):
        self.indexer.abort()
        self.indexer.manager.connection.abort.assert_called_with()

    def test_manager_returns_registered_utility(self):
        self.indexer._manager = None
        util = object()
        provideUtility(util, ISolrConnectionManager)
        self.assertEqual(util, self.indexer.manager)

    def test_index_does_not_call_add_if_disabled(self):
        self.disable_indexing()
        self.indexer.index(object())
        self.assertFalse(
            MockSolrIndexHandler.add.called, 'Should not call add')

    def test_unindex_does_not_call_delete_if_disabled(self):
        self.disable_indexing()
        self.indexer.unindex(object())
        self.assertFalse(
            MockSolrIndexHandler.delete.called, 'Should not call delete')

    def test_commit_does_not_call_commit_if_disabled(self):
        self.disable_indexing()
        self.indexer.commit()
        self.assertFalse(
            self.indexer.manager.connection.commit.called,
            'Should not call commit')

    def test_abort_does_not_call_abort_if_disabled(self):
        self.disable_indexing()
        self.indexer.abort()
        self.assertFalse(
            self.indexer.manager.connection.abort.called,
            'Should not call abort')



# -*- coding: utf-8 -*-
from App.config import getConfiguration
from ftw.solr.converters import to_iso8601
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrIndexHandler
from ftw.solr.interfaces import ISolrSettings
from itertools import islice
from logging import getLogger
from os.path import dirname
from os.path import join as pjoin
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import ICatalogAware
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import base_hasattr
from Products.Five.browser import BrowserView
from time import clock
from time import strftime
from time import time
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.component.hooks import getSite
import logging
from logging.handlers import TimedRotatingFileHandler
import transaction


logger = getLogger('ftw.solr.maintenance')
logger.setLevel(logging.DEBUG)


def get_logfile_path():
    """Determine the path for solr-maintenance.log.
    This will be derived from Zope2's EventLog location, in order to not
    have to figure out the path to var/log/ and the instance name ourselves.
    """
    zconf = getConfiguration()
    eventlog = getattr(zconf, 'eventlog', None)
    if eventlog is None:
        return None

    handler_factories = eventlog.handler_factories
    eventlog_path = handler_factories[0].section.path
    logdir = dirname(eventlog_path)
    return pjoin(logdir, 'solr-maintenance.log')


def setup_maintenance_loghandler():
    logfile_path = get_logfile_path()
    if not logfile_path:
        return

    handler = TimedRotatingFileHandler(
        logfile_path,
        when='midnight', backupCount=7)
    handler.setLevel(logging.DEBUG)
    fmt = "%(asctime)s - %(levelname)s - %(message)s"
    handler.setFormatter(logging.Formatter(fmt))
    logger.addHandler(handler)


setup_maintenance_loghandler()


def timer(func=time):
    """Set up a generator returning the elapsed time since the last call """
    def gen(last=func()):
        while True:
            elapsed = func() - last
            last = func()
            yield '%.3fs' % elapsed
    return gen()


def checkpoint_iterator(function, interval=100):
    """The iterator will call the given function for every nth invocation """
    counter = 0
    while True:
        counter += 1
        if counter % interval == 0:
            function()
        yield None


def solr_date(date):
    """Solr date representation. Fractional seconds are stripped if 0."""
    value = to_iso8601(date)
    if value.endswith('.000Z'):
        value = value[:-5] + 'Z'
    return value


def ellipsified_join(items, max):
    if max is not None:
        items_to_join = list(islice(items, max))
        if len(items) > max:
            items_to_join.append('...')
    else:
        items_to_join = items
    return ', '.join(items_to_join)


class SolrMaintenanceView(BrowserView):
    """Helper view for indexing content in Solr."""

    def __init__(self, context, request):
        super(SolrMaintenanceView, self).__init__(context, request)
        self.request.RESPONSE.setHeader('Content-Type', 'text/plain')

    def is_enabled(self):
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(ISolrSettings)
        return settings.enabled

    def optimize(self):
        """Optimize the Solr index."""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'
        conn = self.manager.connection
        conn.optimize()
        return 'Solr index optimized.'

    def clear(self):
        """Clear all data from Solr index."""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'
        conn = self.manager.connection
        conn.delete_by_query('*:*')
        conn.commit(soft_commit=False)
        return 'Solr index cleared.'

    def reindex(self, commit_interval=100, idxs=None, doom=True):
        """Reindex content in Solr."""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'

        processed = 0
        real = timer()
        lap = timer()
        cpu = timer(clock)

        if doom:
            transaction.doom()

        zodb_conn = self.context._p_jar

        def commit():
            conn = self.manager.connection
            conn.commit(soft_commit=False, extract_after_commit=False)
            zodb_conn.cacheGC()
            self.log(
                'Intermediate commit (%d items processed, last batch in %s)',
                processed, lap.next())

        cpi = checkpoint_iterator(commit, interval=commit_interval)
        self.log('Reindexing Solr...')
        for path, obj in find_objects(self.context):

            if not ICatalogAware.providedBy(obj):
                continue

            handler = getMultiAdapter((obj, self.manager), ISolrIndexHandler)
            handler.add(idxs)
            processed += 1
            cpi.next()

        commit()
        self.log('Solr index rebuilt.')
        self.log(
            'Processed %d items in %s (%s cpu time).',
            processed, real.next(), cpu.next())

    def reindex_cataloged(self, commit_interval=100, idxs=None, start=0,
                          end=-1, query=None, doom=True):
        """Reindex all cataloged content in Solr."""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'

        query = query or {}
        for key, value in self.request.form.items():
            if key in ['UID', 'path', 'created', 'modified', 'portal_type',
                       'object_provides', 'sort_on', 'sort_order']:
                query[key] = value
        if 'sort_on' not in query:
            query['sort_on'] = 'path'
        catalog = getToolByName(self.context, 'portal_catalog')
        items = catalog.unrestrictedSearchResults(**query)

        try:
            start = int(start)
            end = int(end)
            commit_interval = int(commit_interval)
        except ValueError:
            start = 0
            end = -1
            commit_interval = 100

        processed = 0
        real = timer()
        lap = timer()
        cpu = timer(clock)

        if doom:
            transaction.doom()

        zodb_conn = self.context._p_jar

        def commit():
            conn = self.manager.connection
            conn.commit(soft_commit=False, extract_after_commit=False)
            zodb_conn.cacheGC()
            self.log(
                'Intermediate commit (%d items processed, last batch in %s)',
                processed, lap.next())

        cpi = checkpoint_iterator(commit, interval=commit_interval)
        self.log('Reindexing Solr...')
        site = getSite()
        for item in items[start:end]:
            path = item.getPath()
            obj = site.unrestrictedTraverse(path, None)
            if obj is None:
                logger.warning("Object at path %s doesn't exist", path)
                continue

            handler = getMultiAdapter((obj, self.manager), ISolrIndexHandler)
            handler.add(idxs)
            processed += 1
            cpi.next()

        commit()
        self.log('Solr index rebuilt.')
        self.log(
            'Processed %d items in %s (%s cpu time).',
            processed, real.next(), cpu.next())

    def diff(self, max_diff=5):
        """Diff with portal catalog"""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'

        catalog = getToolByName(self.context, 'portal_catalog')
        items = catalog.unrestrictedSearchResults()
        catalog_uids = set([item.UID for item in items])
        catalog_modified = set(
            [(item.UID, solr_date(item.modified)) for item in items])

        conn = self.manager.connection
        resp = conn.search({
            u'query': u'*:*',
            u'limit': 10000000,
            u'params': {u'fl': ['UID', 'modified']},
        })
        solr_uids = set([doc['UID'].encode('utf8') for doc in resp.docs])
        solr_modified = set(
            [(doc['UID'], doc.get('modified', u'2000-01-01T00:00:00.000Z'))
             for doc in resp.docs])

        self.log('Portal Catalog contains %s items.', len(catalog_uids))
        self.log('Solr contains %s items.', len(solr_uids))
        not_in_catalog = solr_uids - catalog_uids
        if not_in_catalog:
            self.log(
                'Total of %s items not in Portal Catalog: %s',
                len(not_in_catalog),
                ellipsified_join(not_in_catalog, max_diff))
        not_in_solr = catalog_uids - solr_uids
        if not_in_solr:
            self.log(
                'Total of %s items not in Solr: %s',
                len(not_in_solr),
                ellipsified_join(not_in_solr, max_diff))
        if not not_in_catalog and not not_in_solr:
            self.log('Solr and Portal Catalog contain the same items. :-)')

        not_in_sync = [item[0] for item in catalog_modified - solr_modified]
        incomplete = conn.search({
            u'query': u'-created:[* TO *]',
            u'limit': 10000000,
            u'params': {u'fl': 'UID'},
        })
        not_in_sync.extend([doc['UID'] for doc in incomplete.docs])
        if not_in_sync:
            self.log(
                'Total of %s items not in sync: %s',
                len(not_in_sync),
                ellipsified_join(not_in_sync, max_diff))
        return not_in_catalog, not_in_solr, not_in_sync

    def sync(self, commit_interval=100, idxs=None, doom=True, max_diff=5):
        """Sync Solr with portal catalog"""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'

        catalog = getToolByName(self.context, 'portal_catalog')
        not_in_catalog, not_in_solr, not_in_sync = self.diff(max_diff=max_diff)

        if not not_in_sync and not not_in_catalog:
            return

        # (Re)index items in Solr that are not in sync
        processed = 0
        real = timer()
        lap = timer()
        cpu = timer(clock)

        if doom:
            transaction.doom()

        zodb_conn = self.context._p_jar

        def commit():
            conn = self.manager.connection
            conn.commit(soft_commit=False, extract_after_commit=False)
            zodb_conn.cacheGC()
            self.log(
                'Intermediate commit (%d items processed, last batch in %s)',
                processed, lap.next())

        cpi = checkpoint_iterator(commit, interval=commit_interval)
        self.log('Syncing Solr...')
        for uid in not_in_sync:
            catalog_item = catalog.unrestrictedSearchResults(UID=uid)[0]
            obj = catalog_item.getObject()

            handler = getMultiAdapter((obj, self.manager), ISolrIndexHandler)
            handler.add(idxs)
            processed += 1
            cpi.next()

        commit()

        # Delete items in Solr that are not in the catalog
        conn = self.manager.connection
        for uid in not_in_catalog:
            conn.delete(uid)
        conn.commit(soft_commit=False)

        self.log('Solr index synced.')
        self.log(
            'Processed %d items in %s (%s cpu time).',
            processed, real.next(), cpu.next())

    def log(self, msg, *args):
        logger.info(msg, *args)
        self.request.RESPONSE.write(
            strftime('%Y-%m-%d %H:%M:%S ') + msg % args + '\n')

    @property
    def manager(self):
        return queryUtility(ISolrConnectionManager)


def find_objects(context):
    """Generator to recursively find and yield all objects below the given
       context."""
    traverse = context.unrestrictedTraverse
    base = '/'.join(context.getPhysicalPath())
    cut = len(base) + 1
    paths = [base]
    for idx, path in enumerate(paths):
        obj = traverse(path)
        yield path[cut:], obj
        if base_hasattr(obj, 'objectIds'):
            for id_ in obj.objectIds():
                paths.insert(idx + 1, path + '/' + id_)



# -*- coding: utf-8 -*-
"""Setup tests for this package."""
from plone import api
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING  # noqa

import unittest


class TestSetup(unittest.TestCase):
    """Test that ftw.solr is properly installed."""

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        """Custom shared utility setup for tests."""
        self.portal = self.layer['portal']
        self.installer = api.portal.get_tool('portal_quickinstaller')

    def test_product_installed(self):
        """Test if ftw.solr is installed."""
        self.assertTrue(self.installer.isProductInstalled(
            'ftw.solr'))

    def test_browserlayer(self):
        """Test that IFtwSolrLayer is registered."""
        from ftw.solr.interfaces import (
            IFtwSolrLayer)
        from plone.browserlayer import utils
        self.assertIn(
            IFtwSolrLayer,
            utils.registered_layers())


class TestUninstall(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.installer = api.portal.get_tool('portal_quickinstaller')
        self.installer.uninstallProducts(['ftw.solr'])

    def test_product_uninstalled(self):
        """Test if ftw.solr is cleanly uninstalled."""
        self.assertFalse(self.installer.isProductInstalled(
            'ftw.solr'))

    def test_browserlayer_removed(self):
        """Test that IFtwSolrLayer is removed."""
        from ftw.solr.interfaces import \
            IFtwSolrLayer
        from plone.browserlayer import utils
        self.assertNotIn(
           IFtwSolrLayer,
           utils.registered_layers())



# -*- coding: utf-8 -*-
from datetime import date
from DateTime import DateTime
from datetime import datetime
from ftw.solr.converters import to_iso8601
from ftw.solr.converters import to_unicode
import pytz
import unittest


class TestDateTimeConverter(unittest.TestCase):

    def test_zope_datetime_in_utc_converts_to_iso8601(self):
        dt = to_iso8601(DateTime('2017-10-21 14:28:16.936207 UTC'))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'2017-10-21T14:28:16.936Z')

    def test_zope_datetime_in_gmt2_converts_to_iso8601(self):
        dt = to_iso8601(DateTime('2017-10-21 16:28:16.936207 GMT+2'))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'2017-10-21T14:28:16.936Z')

    def test_python_datetime_timezone_naive_converts_to_iso8601(self):
        dt = to_iso8601(datetime(2017, 10, 21, 14, 28, 16, 936207))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'2017-10-21T14:28:16.936Z')

    def test_python_datetime_timezone_aware_converts_to_iso8601(self):
        tz = pytz.timezone('Europe/Zurich')
        dt = to_iso8601(tz.localize(
            datetime(2017, 10, 21, 16, 28, 16, 936207)))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'2017-10-21T14:28:16.936Z')

    def test_datetime_before_1900_converts_to_iso8601(self):
        dt = to_iso8601(datetime(1871, 3, 30, 1, 2))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'1871-03-30T01:02:00.000Z')

    def test_datetime_before_1000_converts_to_iso8601(self):
        dt = to_iso8601(datetime(971, 3, 30, 17))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'0971-03-30T17:00:00.000Z')

    def test_datetime_before_100_converts_to_iso8601(self):
        dt = to_iso8601(datetime(71, 3, 30))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'0071-03-30T00:00:00.000Z')

    def test_python_date_converts_to_iso8601(self):
        dt = to_iso8601(date(2017, 10, 21))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'2017-10-21T00:00:00.000Z')

    def test_date_before_1900_converts_to_iso8601(self):
        dt = to_iso8601(date(1871, 3, 30))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'1871-03-30T00:00:00.000Z')

    def test_date_before_1000_converts_to_iso8601(self):
        dt = to_iso8601(date(971, 3, 30))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'0971-03-30T00:00:00.000Z')

    def test_date_before_100_converts_to_iso8601(self):
        dt = to_iso8601(date(71, 3, 30))
        self.assertIsInstance(dt, unicode)
        self.assertEqual(dt, u'0071-03-30T00:00:00.000Z')

    def test_invalid_date_converts_to_none(self):
        self.assertEqual(to_iso8601('30'), None)
        self.assertEqual(to_iso8601(30), None)

    def test_zope_datetime_high_millis_does_not_cause_rounding_error_low(self):
        dt = to_iso8601(DateTime('2017-10-21 16:28:59.999501'))
        self.assertIsInstance(dt, unicode)
        self.assertNotEqual(dt, u'2017-10-21T16:28:60.000Z')
        self.assertEqual(dt, u'2017-10-21T16:28:59.999Z')

    def test_zope_datetime_high_millis_does_not_cause_rounding_error_high(self):
        dt = to_iso8601(DateTime('2017-10-21 16:28:59.999999'))
        self.assertIsInstance(dt, unicode)
        self.assertNotEqual(dt, u'2017-10-21T16:28:60.000Z')
        self.assertEqual(dt, u'2017-10-21T16:28:59.999Z')

    def test_datetime_high_millis_does_not_cause_rounding_error_low(self):
        dt = to_iso8601(datetime(2017, 10, 21, 16, 28, 59, 999501))
        self.assertIsInstance(dt, unicode)
        self.assertNotEqual(dt, u'2017-10-21T16:28:60.000Z')
        self.assertEqual(dt, u'2017-10-21T16:28:59.999Z')

    def test_datetime_high_millis_does_not_cause_rounding_error_high(self):
        dt = to_iso8601(datetime(2017, 10, 21, 16, 28, 59, 999999))
        self.assertIsInstance(dt, unicode)
        self.assertNotEqual(dt, u'2017-10-21T16:28:60.000Z')
        self.assertEqual(dt, u'2017-10-21T16:28:59.999Z')

    def test_zope_and_python_datetimes_get_converted_identically_high_millis(self):
        tz = pytz.timezone('Europe/Zurich')
        python_datetime = tz.localize(datetime(2017, 10, 21, 16, 28, 59, 999999))
        zope_datetime = DateTime(python_datetime)
        self.assertEqual(to_iso8601(python_datetime), to_iso8601(zope_datetime))

    def test_zope_and_python_datetimes_get_converted_identically_low_millis(self):
        tz = pytz.timezone('Europe/Zurich')
        python_datetime = tz.localize(datetime(2017, 10, 21, 16, 28, 59, 111111))
        zope_datetime = DateTime(python_datetime)
        self.assertEqual(to_iso8601(python_datetime), to_iso8601(zope_datetime))


class TestStringConverter(unittest.TestCase):

    def test_bytestring_converts_to_unicode(self):
        string = to_unicode('fünf vor zwölf')
        self.assertIsInstance(string, unicode)
        self.assertEqual(string, u'fünf vor zwölf')

    def test_unicode_converts_to_unicode(self):
        string = to_unicode(u'fünf vor zwölf')
        self.assertIsInstance(string, unicode)
        self.assertEqual(string, u'fünf vor zwölf')

    def test_list_of_bytestrings_converts_to_list_of_unicode(self):
        strings = to_unicode(['fünf', 'vor', 'zwölf'], multivalued=True)
        for string in strings:
            self.assertIsInstance(string, unicode)
        self.assertEqual(strings, [u'fünf', u'vor', u'zwölf'])

    def test_tuple_of_bytestrings_converts_to_list_of_unicode(self):
        strings = to_unicode(('fünf', 'vor', 'zwölf'), multivalued=True)
        for string in strings:
            self.assertIsInstance(string, unicode)
        self.assertEqual(strings, [u'fünf', u'vor', u'zwölf'])

    def test_list_of_bytestrings_converts_to_unicode(self):
        string = to_unicode(['fünf', 'vor', 'zwölf'])
        self.assertIsInstance(string, unicode)
        self.assertEqual(string, u'fünf vor zwölf')

    def test_bytestring_converts_to_list_of_unicode(self):
        string = to_unicode('fünf vor zwölf', multivalued=True)
        self.assertIsInstance(string[0], unicode)
        self.assertEqual(string, [u'fünf vor zwölf'])



from ftw.solr.interfaces import ISolrDocument
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.interface import implementer


@implementer(ISolrDocument)
class SolrDocument(object):

    def __init__(self, data, fields=None):
        self.data = unicode2bytes(data)
        self.fields = fields or list()

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __getattr__(self, name):
        if name in self.data:
            return self.data[name]
        if name in self.fields:
            # Solr response does not include fields that don't have a value
            # But that should not be treated as an AttributeError. It instead
            # means the value for that field is None
            return None
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (
                self.__class__.__name__, name))

    @property
    def getId(self):
        return self.data.get('id')

    def getPath(self):
        return self.path

    def getURL(self, relative=False):
        request = getRequest()
        return request.physicalPathToURL(self.getPath(), relative)

    def getObject(self, REQUEST=None, restricted=True):
        site = getSite()
        path = self.getPath()
        if not path:
            return None
        segments = path.split('/')
        if restricted:
            parent = site.unrestrictedTraverse(segments[:-1], None)
            if parent is None:
                return None
            obj = parent.restrictedTraverse(segments[-1], None)
        else:
            obj = site.unrestrictedTraverse(segments, None)
        # If there's no object at the given path we can still get one through
        # acquisition, but that's not what we want.
        if obj is not None and '/'.join(obj.getPhysicalPath()) == path:
            return obj
        else:
            return None


def unicode2bytes(data):
    if isinstance(data, unicode):
        return data.encode('utf8')
    elif isinstance(data, dict):
        return dict(map(unicode2bytes, data.iteritems()))
    elif isinstance(data, (list, tuple)):
        return list(map(unicode2bytes, data))
    else:
        return data



from ftw.solr.converters import to_iso8601
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrSettings
from logging import getLogger
from plone.registry.interfaces import IRegistry
from zope.component import getUtility
from ZPublisher.HTTPRequest import record

import re

logger = getLogger('ftw.solr.query')

SPECIAL_CHARS = [
    '+', '-', '&&', '||', '!', '(', ')', '{', '}', '[', ']', '^', '"', '~',
    '*', '?', ':', '/']

OPERATORS = re.compile(r'(.*)\s+(AND|OR|NOT)\s+', re.UNICODE)


def escape(term):
    for char in SPECIAL_CHARS:
        term = term.replace(char, '\\' + char)
    return term


def is_simple_search(phrase):
    num_quotes = phrase.count('"')
    if num_quotes % 2 == 0:
        # Replace quoted parts with a marker
        # "foo bar" -> quoted
        parts = phrase.split('"')
        new_parts = []
        for i in range(0, len(parts)):
            if i % 2 == 0:
                new_parts.append(parts[i])
            else:
                new_parts.append('quoted')
        phrase = u''.join(new_parts)
    if bool(OPERATORS.match(phrase)):
        return False
    return True


def split_simple_search(phrase):
    parts = phrase.split('"')
    terms = []
    for i in range(0, len(parts)):
        if i % 2 == 0:
            # Unquoted text
            terms.extend([term for term in parts[i].split() if term])
        else:
            # The uneven parts are those inside quotes
            if parts[i]:
                terms.append('"%s"' % parts[i])
    return terms


def make_query(phrase):
    phrase = phrase.strip()
    if isinstance(phrase, str):
        phrase = phrase.decode('utf8')
    registry = getUtility(IRegistry)
    settings = registry.forInterface(ISolrSettings)
    if is_simple_search(phrase):
        terms = split_simple_search(phrase)[:10]
        pattern = settings.simple_search_term_pattern
        term_queries = [pattern.format(term=escape(t)) for t in terms]
        if len(term_queries) > 1:
            term_queries = [u'(%s)' % q for q in term_queries]
        query = u' AND '.join(term_queries)
        if len(terms) > 1 or not phrase.isalnum():
            query = u'(%s) OR (%s)' % (
                settings.simple_search_phrase_pattern.format(
                    phrase=escape(phrase)),
                query,
            )
    else:
        pattern = settings.complex_search_pattern
        query = pattern.format(phrase=phrase)
    if settings.local_query_parameters:
        query = settings.local_query_parameters + query
    return query


def make_path_filter(path, depth=0, include_self=False):
    filters = []
    if depth == 0:
        filters.append(u'path:{}'.format(ensure_text(escape(path))))
    else:
        filters.append(u'path_parent:{}'.format(ensure_text(escape(path))))
    if depth > 0:
        current_depth = len(path.split('/')) - 1
        min_depth = current_depth if include_self else current_depth + 1
        max_depth = current_depth + depth
        if max_depth == min_depth:
            filters.append(u'path_depth:{}'.format(max_depth))
        else:
            filters.append(u'path_depth:[{} TO {}]'.format(
                min_depth, max_depth))
    return filters


def make_filters(**kwargs):
    manager = getUtility(ISolrConnectionManager)
    filters = []
    for key, value in kwargs.items():
        if key not in manager.schema.fields:
            logger.warning(
                'Ignoring filter criteria for unknown field %s', key)
            continue
        elif key == 'path':
            if isinstance(value, dict):
                filters.extend(make_path_filter(
                    value.get('query'), depth=value.get('depth', 0)))
            else:
                filters.extend(make_path_filter(value))
        elif isinstance(value, (list, tuple)):
            filters.append(u'{}:({})'.format(
                key, escape(u' OR '.join(ensure_text(value)))))
        elif isinstance(value, bool):
            filters.append(u'{}:{}'.format(key, u'true' if value else u'false'))
        elif isinstance(value, (dict, record)):
            query = value.get('query')
            operator = value.get('operator')
            range_ = value.get('range', None)
            if query and isinstance(query, (list, tuple)) and operator:
                operator = u' {} '.format(operator.upper())
                filters.append(u'{}:({})'.format(
                    key, escape(operator.join(ensure_text(query)))))
            elif query and range_ in ['min', 'max', 'minmax']:
                if not isinstance(query, (list, tuple)):
                    query = [query]
                if range_ == 'min':
                    filters.append(u'{}:[{} TO *]'.format(
                        key, escape(to_iso8601(query[0]))))
                elif range_ == 'max':
                    filters.append(u'{}:[* TO {}]'.format(
                        key, escape(to_iso8601(query[0]))))
                elif range_ == 'minmax' and len(query) > 1:
                    filters.append(u'{}:[{} TO {}]'.format(
                        key,
                        escape(to_iso8601(min(query))),
                        escape(to_iso8601(max(query))),
                    ))
        else:
            filters.append(u'{}:{}'.format(key, escape(ensure_text(value))))
    return filters


def ensure_text(value):
    if isinstance(value, bytes):
        value = value.decode('utf8')
    elif isinstance(value, (int, float)):
        value = unicode(value)
    elif isinstance(value, (list, tuple)):
        value = [ensure_text(v) for v in value]
    return value



# -*- coding: utf-8 -*-
from zope.interface import Interface
from zope import schema
from zope.component.zcml import utility
from ftw.solr.interfaces import ISolrConnectionConfig
from ftw.solr.connection import SolrConnectionConfig


class ISolrConnectionConfigDirective(Interface):
    """Directive which registers a Solr connection config"""

    host = schema.ASCIILine(
        title=u"Host",
        description=u"The host name of the Solr instance to be used.",
        required=True,
    )

    port = schema.Int(
        title=u"Port",
        description=u"The port of the Solr instance to be used.",
        required=True,
    )

    base = schema.ASCIILine(
        title=u"Base",
        description=u"The base prefix of the Solr instance to be used "
                    "(e.g. /solr/corename or /api/cores/corename)",
        required=True,
    )

    upload_blobs = schema.Bool(
        title=u"Upload Blobs",
        description=u"Upload blobs to extract handler via HTTP POST instead "
                    u"of making Solr retrieve them via filesystem.",
        default=False,
    )


def solr_connection_config_directive(_context, host, port, base, upload_blobs):

    utility(_context,
            provides=ISolrConnectionConfig,
            component=SolrConnectionConfig(host, port, base, upload_blobs))



# -*- coding: utf-8 -*-
from DateTime import DateTime
from ftw.solr.connection import SolrResponse
from ftw.solr.handlers import ATBlobFileIndexHandler
from ftw.solr.handlers import DefaultIndexHandler
from ftw.solr.handlers import DexterityItemIndexHandler
from ftw.solr.schema import SolrSchema
from ftw.solr.testing import FTW_SOLR_AT_INTEGRATION_TESTING
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING
from ftw.solr.tests.utils import get_data
from ftw.solr.tests.utils import normalize_whitespaces
from mock import MagicMock
from mock import PropertyMock
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.namedfile.file import NamedBlobFile
from plone.uuid.interfaces import IMutableUUID
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import getFSVersionTuple
import unittest


class TestDefaultIndexHandler(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

        self.doc = api.content.create(
            type='Document', title='My Document', id='doc',
            container=self.portal)
        if base_hasattr(self.doc, '_setUID'):
            self.doc._setUID('09baa75b67f44383880a6dab8b3200b6')
            self.doc.setModificationDate(DateTime('2017-01-21T17:18:19+00:00'))
        else:
            IMutableUUID(self.doc).set('09baa75b67f44383880a6dab8b3200b6')
            self.doc.modification_date = DateTime('2017-01-21T17:18:19+00:00')

        conn = MagicMock(name='SolrConnection')
        conn.get = MagicMock(name='get', return_value=SolrResponse(
            body=get_data('schema.json'), status=200))
        self.manager = MagicMock(name='SolrConnectionManager')
        type(self.manager).connection = PropertyMock(return_value=conn)
        type(self.manager).schema = PropertyMock(return_value=SolrSchema(
            self.manager))
        self.handler = DefaultIndexHandler(self.doc, self.manager)

    def test_get_data_without_attributes_gets_all_fields(self):
        data = self.handler.get_data(None)
        data['SearchableText'] = normalize_whitespaces(data['SearchableText'])
        self.assertEqual(
            {
                u'Title': u'My Document',
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
                u'allowedRolesAndUsers': [u'Anonymous'],
                u'path': u'/plone/doc',
                u'path_depth': 2,
                u'modified': u'2017-01-21T17:18:19.000Z',
                u'SearchableText': 'doc My Document',
            },
            data
        )

    def test_get_data_gets_only_specified_attributes(self):
        self.assertEqual(
            {
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
                u'modified': u'2017-01-21T17:18:19.000Z',
            },
            self.handler.get_data(['UID', 'modified'])
        )

    def test_get_data_always_includes_unique_key(self):
        self.assertEqual(
            {
                u'Title': u'My Document',
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            },
            self.handler.get_data(['Title'])
        )

    def test_get_data_includes_path_depth_if_exists_and_path_was_included(self):
        self.assertEqual(
            {
                u'path': u'/plone/doc',
                u'path_depth': 2,
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            },
            self.handler.get_data(['path'])
        )

        # solr schemas without the path_depth field
        self.manager.schema.fields.pop('path_depth')
        self.assertEqual(
            {
                u'path': u'/plone/doc',
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            },
            self.handler.get_data(['path'])
        )

    def test_add_without_attributes_adds_full_documemt(self):
        self.manager.connection.add = MagicMock(name='add')
        self.handler.add(None)
        data = self.manager.connection.add.call_args[0][0]
        data['SearchableText'] = normalize_whitespaces(data['SearchableText'])
        self.assertEqual(
            {
                u'UID': u'09baa75b67f44383880a6dab8b3200b6',
                u'Title': u'My Document',
                u'modified': u'2017-01-21T17:18:19.000Z',
                u'SearchableText': 'doc My Document',
                u'allowedRolesAndUsers': [u'Anonymous'],
                u'path': u'/plone/doc',
                u'path_depth': 2,
            },
            data
        )

    def test_add_with_attributes_does_atomic_update(self):
        self.manager.connection.add = MagicMock(name='add')
        self.handler.add(['Title', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {'set': u'My Document'},
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })

    def test_add_with_attributes_without_data_does_nothing(self):
        self.manager.connection.add = MagicMock(name='add')
        self.handler.add(['field_not_in_schema'])
        self.assertFalse(self.manager.connection.add.called)

    def test_delete(self):
        self.manager.connection.delete = MagicMock(name='delete')
        self.handler.delete()
        self.manager.connection.delete.assert_called_once_with(
            u'09baa75b67f44383880a6dab8b3200b6')


@unittest.skipIf(getFSVersionTuple() >= (5, 0),
                 'Files are not AT-based on Plone 5 and later')
class TestATBlobFileIndexHandler(unittest.TestCase):

    layer = FTW_SOLR_AT_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

        self.doc = api.content.create(
            type='File', title='My File', id='doc',
            file='File data ...', container=self.portal)

        field = self.doc.getPrimaryField().get(self.doc)
        field.setFilename('document.docx')
        field.setContentType('application/vnd.openxmlformats-officedocument.wordprocessingml.document')

        if base_hasattr(self.doc, '_setUID'):
            self.doc._setUID('09baa75b67f44383880a6dab8b3200b6')
            self.doc.setModificationDate(DateTime('2017-01-21T17:18:19+00:00'))
        else:
            IMutableUUID(self.doc).set('09baa75b67f44383880a6dab8b3200b6')
            self.doc.modification_date = DateTime('2017-01-21T17:18:19+00:00')

        conn = MagicMock(name='SolrConnection')
        conn.get = MagicMock(name='get', return_value=SolrResponse(
            body=get_data('schema.json'), status=200))
        self.manager = MagicMock(name='SolrConnectionManager')
        type(self.manager).connection = PropertyMock(return_value=conn)
        type(self.manager).schema = PropertyMock(return_value=SolrSchema(
            self.manager))
        self.handler = ATBlobFileIndexHandler(self.doc, self.manager)

    def test_add_without_attributes_calls_add_and_extract(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')
        self.handler.add(None)
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {u'set': u'My File'},
            u'modified': {u'set': u'2017-01-21T17:18:19.000Z'},
            u'allowedRolesAndUsers': {u'set': [u'Anonymous']},
            u'path': {u'set': u'/plone/doc'},
            u'path_depth': {u'set': 2},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.getFile().blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_attributes_without_searchabletext_calls_add(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')
        self.handler.add(['Title', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {'set': u'My File'},
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_not_called()

    def test_add_with_attributes_with_searchabletext_calls_add_and_extract(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')
        self.handler.add(['SearchableText', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.getFile().blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_searchabletext_only_calls_extract(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')
        self.handler.add(['SearchableText'])
        self.manager.connection.add.assert_not_called()
        self.manager.connection.extract.assert_called_once_with(
            self.doc.getFile().blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_attributes_without_data_does_nothing(self):
        self.manager.connection.add = MagicMock(name='add')
        self.handler.add(['field_not_in_schema'])
        self.assertFalse(self.manager.connection.add.called)

    def test_add_with_attributes_as_tuple_with_searchabletext(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')

        self.handler.add(('SearchableText', 'modified'))
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.getFile().blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_only_searchable_text_and_uid_calls_extract(self):
        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')

        self.handler.add(('SearchableText', 'UID'))
        self.manager.connection.add.assert_not_called()
        self.manager.connection.extract.assert_called_once_with(
            self.doc.getFile().blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )


class TestDexterityItemIndexHandler(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ['Manager'])

        namedfile = NamedBlobFile(
            data='File data ...',
            filename=u'document.docx',
            contentType='application/vnd.openxmlformats-officedocument.wordprocessingml.document')
        self.doc = api.content.create(
            type='File', title='My File', id='doc',
            file=namedfile, container=self.portal)
        if base_hasattr(self.doc, '_setUID'):
            self.doc._setUID('09baa75b67f44383880a6dab8b3200b6')
            self.doc.setModificationDate(DateTime('2017-01-21T17:18:19+00:00'))
        else:
            IMutableUUID(self.doc).set('09baa75b67f44383880a6dab8b3200b6')
            self.doc.modification_date = DateTime('2017-01-21T17:18:19+00:00')

        conn = MagicMock(name='SolrConnection')
        conn.get = MagicMock(name='get', return_value=SolrResponse(
            body=get_data('schema.json'), status=200))
        self.manager = MagicMock(name='SolrConnectionManager')
        type(self.manager).connection = PropertyMock(return_value=conn)
        type(self.manager).schema = PropertyMock(return_value=SolrSchema(
            self.manager))
        self.handler = DexterityItemIndexHandler(self.doc, self.manager)

        self.manager.connection.add = MagicMock(name='add')
        self.manager.connection.extract = MagicMock(name='extract')

    def test_add_without_attributes_calls_add_and_extract(self):
        self.handler.add(None)
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {u'set': u'My File'},
            u'modified': {u'set': u'2017-01-21T17:18:19.000Z'},
            u'allowedRolesAndUsers': {u'set': [u'Anonymous']},
            u'path': {u'set': u'/plone/doc'},
            u'path_depth': {u'set': 2},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_empty_list_attributes_calls_add_and_extract(self):
        self.handler.add([])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {u'set': u'My File'},
            u'modified': {u'set': u'2017-01-21T17:18:19.000Z'},
            u'allowedRolesAndUsers': {u'set': [u'Anonymous']},
            u'path': {u'set': u'/plone/doc'},
            u'path_depth': {u'set': 2},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_attributes_without_searchabletext_calls_add(self):
        self.handler.add(['Title', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'Title': {'set': u'My File'},
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_not_called()

    def test_add_with_attributes_with_searchabletext_calls_add_and_extract(self):
        self.handler.add(['SearchableText', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_searchabletext_only_calls_extract(self):
        self.handler.add(['SearchableText'])
        self.manager.connection.add.assert_not_called()
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_without_attributes_for_item_without_blob_calls_add(self):
        item = api.content.create(
            type='Document', title='My Document', id='doc2',
            container=self.portal)
        IMutableUUID(item).set('56d2c8b62c064f38890c60b66d01c894')
        item.modification_date = DateTime('2017-01-21T17:18:19+00:00')

        handler = DexterityItemIndexHandler(item, self.manager)
        handler.add(None)
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'56d2c8b62c064f38890c60b66d01c894',
            u'Title': {u'set': u'My Document'},
            u'modified': {u'set': u'2017-01-21T17:18:19.000Z'},
            u'SearchableText': {'set': ' doc2 My Document   '},
            u'allowedRolesAndUsers': {u'set': [u'Anonymous']},
            u'path': {u'set': u'/plone/doc2'},
            u'path_depth': {u'set': 2},
        })
        self.manager.connection.extract.assert_not_called()

    def test_add_with_attributes_for_item_without_blob_calls_add(self):
        item = api.content.create(
            type='Document', title='My Document', id='doc2',
            container=self.portal)
        IMutableUUID(item).set('56d2c8b62c064f38890c60b66d01c894')
        item.modification_date = DateTime('2017-01-21T17:18:19+00:00')

        handler = DexterityItemIndexHandler(item, self.manager)
        handler.add(['SearchableText', 'modified'])
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'56d2c8b62c064f38890c60b66d01c894',
            u'modified': {u'set': u'2017-01-21T17:18:19.000Z'},
            u'SearchableText': {'set': ' doc2 My Document   '},
        })
        self.manager.connection.extract.assert_not_called()

    def test_add_with_attributes_without_data_does_nothing(self):
        self.manager.connection.add = MagicMock(name='add')
        self.handler.add(['field_not_in_schema'])
        self.assertFalse(self.manager.connection.add.called)

    def test_add_with_attributes_as_tuple_with_searchabletext(self):
        self.handler.add(('SearchableText', 'modified'))
        self.manager.connection.add.assert_called_once_with({
            u'UID': u'09baa75b67f44383880a6dab8b3200b6',
            u'modified': {'set': u'2017-01-21T17:18:19.000Z'},
        })
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )

    def test_add_with_only_searchable_text_and_uid_calls_extract(self):
        self.handler.add(('SearchableText', 'UID'))
        self.manager.connection.add.assert_not_called()
        self.manager.connection.extract.assert_called_once_with(
            self.doc.file._blob,
            'SearchableText',
            {u'UID': u'09baa75b67f44383880a6dab8b3200b6'},
            'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
        )



from ftw.upgrade import UpgradeStep


class AddEnableSetting(UpgradeStep):
    """Add registry setting for enabling/disabling indexing
    """

    def __call__(self):
        self.install_upgrade_profile()



from ftw.solr.helpers import chunked_file_reader
from ftw.solr.helpers import group_by_two
from ftw.solr.helpers import http_chunked_encoder
from ftw.solr.interfaces import ISolrConnectionConfig
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.schema import SolrSchema
from httplib import HTTPConnection
from httplib import HTTPException
from logging import getLogger
from threading import local
from urllib import urlencode
from zope.component import queryUtility
from zope.interface import implementer
import json
import os.path
import socket
import transaction


logger = getLogger('ftw.solr.connection')

# a thread-local object holding the connection
local_data = local()


@implementer(ISolrConnectionConfig)
class SolrConnectionConfig(object):

    def __init__(self, host, port, base, upload_blobs=False):
        self.host = host
        self.port = port
        self.base = base
        self.upload_blobs = upload_blobs


class SolrConnection(object):

    post_headers = {'Content-Type': 'application/json'}

    def __init__(self, host='localhost', port=8983, base='/solr',
                 timeout=None, upload_blobs=False):
        self.timeout = timeout
        self.host = host
        self.port = port
        self.base = base
        self.upload_blobs = upload_blobs
        self.conn = HTTPConnection(self.host, self.port, timeout=self.timeout)
        self.update_commands = []
        self.extract_commands = []
        self.reconnect_before_request = False

    def request(self, method, path, data=None, headers={}, log_error=True):
        try:
            if self.reconnect_before_request:
                self.reconnect()
            self.conn.request(
                method, self.base + path, body=data, headers=headers)
            resp = self.conn.getresponse()
            body = resp.read()
            status = resp.status
            self.reconnect_before_request = False
            return SolrResponse(body, status, log_error=log_error)
        except (socket.error, HTTPException) as exc:
            if not self.reconnect_before_request:
                self.reconnect_before_request = True
                return self.request(method, path, data=data, headers=headers)
            else:
                self.reconnect_before_request = False
                return SolrResponse(exception=exc, log_error=log_error)

    def post(self, path, data=None, headers={}, log_error=True):
        headers = headers if headers else self.post_headers
        return self.request(
            'POST', path, data, headers, log_error=log_error)

    def post_chunked(self, path, blob, content_type=None, log_error=True):
        if not content_type:
            content_type = 'application/octet-stream'

        try:
            if self.reconnect_before_request:
                self.reconnect()

            self.conn.putrequest('POST', self.base + path)
            self.conn.putheader('Transfer-Encoding', 'chunked')
            self.conn.putheader('Content-Type', content_type)
            self.conn.endheaders()

            with open(blob.committed(), 'rb') as file_:
                reader = chunked_file_reader(file_)
                for chunk in http_chunked_encoder(reader):
                    self.conn.send(chunk)

            resp = self.conn.getresponse()
            body = resp.read()
            status = resp.status
            self.reconnect_before_request = False
            return SolrResponse(body, status, log_error=log_error)

        except (socket.error, HTTPException) as exc:
            if not self.reconnect_before_request:
                self.reconnect_before_request = True
                return self.post_chunked(path, blob, content_type=content_type)
            else:
                self.reconnect_before_request = False
                return SolrResponse(exception=exc, log_error=log_error)

    def get(self, path, headers={}, log_error=True):
        return self.request('GET', path, headers=headers, log_error=log_error)

    def reconnect(self):
        self.conn.close()
        self.conn.connect()

    def add(self, data):
        self.update_commands.append('"add": ' + json.dumps({'doc': data}))

    def extract(self, blob, field, data, content_type):
        """Add blob using Solr's Extracting Request Handler."""
        self.extract_commands.append((blob, field, data, content_type))

    def delete(self, id_):
        self.update_commands.append(
            '"delete": ' + json.dumps({'id': id_}))

    def delete_by_query(self, query):
        self.update_commands.append(
            '"delete": ' + json.dumps({'query': query}))

    def commit(self, wait_searcher=True, soft_commit=True,
               extract_after_commit=True):
        self.update_commands.append(
            '"commit": ' + json.dumps(
                {'waitSearcher': wait_searcher, 'softCommit': soft_commit}))
        self.flush(extract_after_commit=extract_after_commit)

    def optimize(self, wait_searcher=True):
        self.update_commands.append(
            '"optimize": ' + json.dumps({'waitSearcher': wait_searcher}))
        self.flush()

    def filter_extract_commands(self):
        """We filter the extract commands to only keep the last one
        for each field and UID. The others are not necessary as their result
        would anyway be overwritten by a later extract command. Moreover it
        happens that earlier commands point to outdated blobs, leading to
        errors
        """
        added = set()
        filtered_extract_commands = []
        for command in reversed(self.extract_commands):
            blob, field, data, content_type = command
            uid = data.get("UID")
            if uid is None:
                # This should not happen, but better safe than sorry
                filtered_extract_commands.append(command)
            if (uid, field) in added:
                continue
            filtered_extract_commands.append(command)
            added.add((uid, field))
        self.extract_commands = reversed(filtered_extract_commands)

    def flush(self, extract_after_commit=True):
        """Send queued update commands to Solr."""
        if self.update_commands:
            data = '{%s}' % ','.join(self.update_commands)
            resp = self.post('/update', data=data, log_error=False)
            if not resp.is_ok():
                logger.error('Update command failed. %s', resp.error_msg())
            self.update_commands = []

        if self.extract_commands:
            def hook(succeeded, extract_commands):
                if not succeeded:
                    return
                for blob, field, data, content_type in extract_commands:
                    file_ = blob.committed()
                    params = {}
                    params['extractOnly'] = 'true'

                    if self.upload_blobs:
                        # Upload blobs to extract handler via POST
                        resp = self.post_chunked(
                            '/update/extract?%s' % urlencode(params, doseq=True),
                            blob,
                            content_type=content_type,
                            log_error=False)
                        extracted_fulltext_key = ''
                    else:
                        # Make extract handler retrieve blobs via filesystem
                        params['stream.file'] = file_
                        resp = self.post(
                            '/update/extract?%s' % urlencode(params, doseq=True),
                            headers={'Content-Type': 'application/x-www-form-urlencoded'},  # noqa
                            log_error=False)
                        extracted_fulltext_key = os.path.basename(file_)

                    if not resp.is_ok():
                        logger.error(
                            'Extract command for UID=%s with blob %s failed. '
                            '%s',
                            data.get(u'UID'), file_, resp.error_msg())
                        continue

                    update_command = {"add": {"doc": data}}
                    update_command["add"]["doc"][field] = {
                        "set": resp.body.get(extracted_fulltext_key),
                    }
                    resp = self.post(
                        '/update',
                        data=json.dumps(update_command),
                        log_error=False)
                    if not resp.is_ok():
                        logger.error(
                            'Update command failed. %s', resp.error_msg())

            self.filter_extract_commands()
            if extract_after_commit:
                transaction.get().addAfterCommitHook(
                    hook, args=[self.extract_commands])
            else:
                hook(True, self.extract_commands)
            self.extract_commands = []

    def abort(self):
        """Abort all pending commands."""
        # While Solr supports rollback we can't use it because there could be
        # other uncommitted operations from other connections and those would
        # be rollbacked too. Thus we delay sending anything to Solr until the
        # next commit.
        self.update_commands = []
        self.extract_commands = []

    def search(self, params, request_handler='/select'):
        return self.post(request_handler, json.dumps(params))

    def __repr__(self):
        return "SolrConnection(host='%s', port=%s, base='%s') at 0x%x" % (
            self.host, self.port, self.base, id(self))


_no_connection_marker = object()


@implementer(ISolrConnectionManager)
class SolrConnectionManager(object):

    @property
    def connection(self):
        conn = getattr(local_data, 'connection', _no_connection_marker)
        if conn is _no_connection_marker:
            config = queryUtility(ISolrConnectionConfig)
            if config is not None:
                conn = SolrConnection(
                    host=config.host, port=config.port, base=config.base,
                    upload_blobs=config.upload_blobs)
            else:
                conn = None
                logger.warning('Solr configuration missing.')
            setattr(local_data, 'connection', conn)
        return conn

    @property
    def schema(self):
        schema = getattr(local_data, 'schema', None)
        if not schema:
            schema = SolrSchema(manager=self)
            setattr(local_data, 'schema', schema)
        return schema


class SolrResponse(object):

    def __init__(self, body='null', status=None, exception=None,
                 log_error=True):
        self.http_status = status
        self.exception = exception
        self.status = -1
        self.qtime = 0
        self.params = None
        self.docs = []
        self.facets = {}
        self.num_found = 0
        self.start = 0
        self.body = {}
        if self.http_status:
            self.parse(body)
        if log_error and not self.is_ok():
            logger.error('Solr response error. %s', self.error_msg())

    def parse(self, data):
        try:
            data = json.loads(data)
        except ValueError:
            data = None
        if data is None:
            return

        response_header = data.get(u'responseHeader', {})
        self.status = response_header.get(u'status', self.status)
        self.qtime = response_header.get(u'QTime', 0)
        self.params = json.loads(response_header.get(u'params', {}).get(
            u'json', 'null'))

        if u'response' in data:
            self.docs = data[u'response'].get(u'docs', [])
            self.num_found = data[u'response'].get(u'numFound', 0)
            self.start = data[u'response'].get(u'start', 0)

        if u'facet_counts' in data:
            facet_fields = data[u'facet_counts'].get(u'facet_fields', {})
            for field, counts in facet_fields.items():
                # facet counts are arranged in a tuple with the form
                # (facet1, count1, facet2, count2, ...)
                self.facets[field] = dict(group_by_two(counts))
        self.body = data

    def is_ok(self):
        if self.http_status == 200 and self.status == 0:
            return True
        return False

    def error_msg(self):
        if self.exception is not None:
            return 'An exception occured: %s' % str(self.exception)
        elif self.http_status:
            if u'error' in self.body and u'msg' in self.body[u'error']:
                return 'Server responded with code %s, %s.' % (
                    self.http_status,
                    self.body[u'error'][u'msg'].encode('utf8'))
            else:
                return 'Server responded with code %s.' % self.http_status
        else:
            return 'An unknown error occured.'

    def get(self, key, default=None):
        return self.body.get(key, default)

    def __getitem__(self, key):
        return self.body[key]

    def __contains__(self, key):
        return key in self.body

    def __repr__(self):
        if self.is_ok():
            return 'SolrResponse(%s docs)' % len(self.docs)
        elif self.exception is not None:
            return 'SolrResponse(exception=%r)' % self.exception
        return "SolrResponse(status=%s)" % self.http_status



from Products.CMFCore.utils import getToolByName
from ftw.solr.converters import CONVERTERS
from ftw.solr.interfaces import ISolrIndexHandler
from logging import getLogger
from plone.indexer.interfaces import IIndexableObject
from plone.namedfile.interfaces import INamedBlobFile
from plone.rfc822.interfaces import IPrimaryFieldInfo
from zope.component import queryMultiAdapter
from zope.interface import implementer


logger = getLogger('ftw.solr.handlers')


@implementer(ISolrIndexHandler)
class DefaultIndexHandler(object):

    def __init__(self, context, manager):
        self.context = context
        self.manager = manager

    def add(self, attributes):
        if self.manager.connection is None:
            return

        error = self.get_schema_error()
        if error:
            logger.warning('%s, skipping indexing of %r', error, self.context)
            return

        data = self.get_data(attributes)

        unique_key = self.manager.schema.unique_key
        if unique_key not in data:
            logger.warning(
                'Object is missing unique key, skipping indexing of %r',
                self.context)
            return

        if attributes:
            data = self.add_atomic_update_modifier(data, unique_key)

        if data:
            self.manager.connection.add(data)

    def delete(self):
        if self.manager.connection is None:
            return

        error = self.get_schema_error()
        if error:
            logger.warning(
                '%s, skipping unindexing of %r', error, self.context)
            return

        unique_key = self.manager.schema.unique_key
        data = self.get_data([unique_key])
        if unique_key not in data:
            logger.warning(
                'Object is missing unique key, skipping unindexing of %r',
                self.context)
            return

        self.manager.connection.delete(data[unique_key])

    def get_data(self, attributes):
        if not IIndexableObject.providedBy(self.context):
            catalog = getToolByName(self.context, 'portal_catalog')
            wrapped = queryMultiAdapter(
                (self.context, catalog), IIndexableObject,
                default=self.context)
        else:
            wrapped = self.context

        schema = self.manager.schema
        if not attributes:
            attributes = set(schema.fields.keys())
        else:
            attributes = set(schema.fields.keys()).intersection(attributes)

        if schema.unique_key not in attributes:
            attributes.add(schema.unique_key)

        # If path gets (re)indexed, we also need to (re)index path_depth
        # in Solr. Because path_depth is not a catalog index, we would
        # otherwise fail to update it in cases where a specific list of
        # `idxs` including path is passed to reindexObject().
        if 'path' in attributes and 'path_depth' in schema.fields:
            attributes.add('path_depth')

        data = {}
        for name in attributes:
            if name.startswith('_'):
                continue
            if name == 'path':
                value = '/'.join(self.context.getPhysicalPath())
            elif name == 'path_depth':
                value = len(self.context.getPhysicalPath()) - 1
            else:
                try:
                    value = getattr(wrapped, name)
                    if callable(value):
                        value = value()
                except (AttributeError, TypeError):
                    continue

            field_class = schema.field_types[
                schema.fields[name][u'type']][u'class']
            multivalued = schema.fields[name].get(u'multiValued', False)
            converter = CONVERTERS.get(field_class)
            if converter is not None:
                value = converter(value, multivalued)
            data[name] = value

        return data

    def get_schema_error(self):
        if not self.manager.schema:
            return 'Unable to fetch schema'

        if self.manager.schema.unique_key is None:
            return 'Schema is missing unique key'

    def add_atomic_update_modifier(self, data, unique_key):
        id_ = data[unique_key]
        data = {k: {'set': v} for k, v in data.items() if k != unique_key}
        if not data:
            return
        data[unique_key] = id_
        return data


@implementer(ISolrIndexHandler)
class ATBlobFileIndexHandler(DefaultIndexHandler):

    def add(self, attributes):
        if self.manager.connection is None:
            return

        if isinstance(attributes, tuple):
            attributes = list(attributes)

        error = self.get_schema_error()
        if error:
            logger.warning('%s, skipping indexing of %r', error, self.context)
            return

        if attributes is None:
            attributes = self.manager.schema.fields.keys()

        extract = False
        if 'SearchableText' in attributes:
            attributes.remove('SearchableText')
            extract = True

        unique_key = self.manager.schema.unique_key
        raw_data = self.get_data(attributes)

        if attributes:
            data = self.add_atomic_update_modifier(raw_data, unique_key)
            if data:
                self.manager.connection.add(data)

        if extract:
            field = self.context.getPrimaryField()
            blob = field.get(self.context).blob
            content_type = field.get(self.context).getContentType()
            self.manager.connection.extract(
                blob, 'SearchableText', {unique_key: raw_data[unique_key]},
                content_type)


@implementer(ISolrIndexHandler)
class DexterityItemIndexHandler(DefaultIndexHandler):

    def add(self, attributes):
        if self.manager.connection is None:
            return

        if isinstance(attributes, tuple):
            attributes = list(attributes)

        error = self.get_schema_error()
        if error:
            logger.warning('%s, skipping indexing of %r', error, self.context)
            return

        blob = None
        content_type = None
        try:
            info = IPrimaryFieldInfo(self.context, None)
        except TypeError:
            info = None
        if info is not None and INamedBlobFile.providedBy(info.value):
            blob = info.value._blob
            content_type = info.value.contentType

        if not attributes:
            attributes = self.manager.schema.fields.keys()

        extract = False
        if 'SearchableText' in attributes and blob is not None:
            attributes.remove('SearchableText')
            extract = True

        unique_key = self.manager.schema.unique_key
        raw_data = self.get_data(attributes)

        if attributes:
            data = self.add_atomic_update_modifier(raw_data, unique_key)
            if data:
                self.manager.connection.add(data)

        if extract:
            self.manager.connection.extract(
                blob, 'SearchableText', {unique_key: raw_data[unique_key]},
                content_type)



from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SpecialUsers import system as system_user
from Products.CMFPlone.interfaces import IPloneSiteRoot
from Testing.makerequest import makerequest
from zope.component import queryMultiAdapter
from zope.component.hooks import setSite
import argparse
import sys


def setup_site(app, options):
    # If no plone site was provided by the command line, try to find one.
    if options.site is None:
        sites = get_plone_sites(app)
        if len(sites) == 1:
            portal = sites[0]
        elif len(sites) > 1:
            sys.exit("Multiple Plone sites found. Please specify which Plone "
                     "site should be used.")
        else:
            sys.exit("No Plone site found.")
    else:
        portal = app.unrestrictedTraverse(options.site, None)
    if not portal:
        sys.exit("Plone site not found at %s" % options.plone_site)

    user = system_user.__of__(app.acl_users)
    newSecurityManager(app, user)

    setSite(portal)
    return portal


def get_plone_sites(root):
    result = []
    for obj in root.values():
        if obj.meta_type == 'Folder':
            result = result + get_plone_sites(obj)
        elif IPloneSiteRoot.providedBy(obj):
            result.append(obj)
        elif obj.getId() in getattr(root, '_mount_points', {}):
            result.extend(get_plone_sites(obj))
    return result


def solr(app, args):
    parser = argparse.ArgumentParser()
    parser.add_argument('command', help='Command to perform',
                        choices=['reindex', 'reindex-cataloged', 'sync',
                                 'optimize', 'clear', 'diff'])
    parser.add_argument('-s', dest='site', default=None,
                        help='Absolute path to the Plone site')
    parser.add_argument('--commit-interval', dest='commit_interval',
                        default=100, type=int,
                        help='Interval for intermediate commits')
    parser.add_argument('-i', '--indexes', nargs='*',
                        help='Perform command on the given indexes')
    parser.add_argument('--max-diff', dest='max_diff', default=5, type=int,
                        help="Maximum items to log in diff. Use negative "
                             "number for infinite.")
    options = parser.parse_args(args[2:])
    app = makerequest(app)
    site = setup_site(app, options)

    max_diff = options.max_diff
    if max_diff < 0:
        max_diff = None

    solr_maintenance = queryMultiAdapter(
        (site, site.REQUEST), name=u'solr-maintenance')
    if options.command == 'reindex':
        solr_maintenance.reindex(
            commit_interval=options.commit_interval, idxs=options.indexes)
    elif options.command == 'reindex-cataloged':
        solr_maintenance.reindex_cataloged(
            commit_interval=options.commit_interval, idxs=options.indexes)
    elif options.command == 'sync':
        solr_maintenance.sync(
            commit_interval=options.commit_interval, idxs=options.indexes,
            max_diff=max_diff)
    elif options.command == 'optimize':
        solr_maintenance.optimize()
        print("Solr index optimized.")
    elif options.command == 'clear':
        solr_maintenance.clear()
        print("Solr index cleared.")
    elif options.command == 'diff':
        solr_maintenance.diff(max_diff=max_diff)
    else:
        sys.exit("Unknown command '%s'." % options.command)



# -*- coding: utf-8 -*-
from ftw.solr.contentlisting import SolrDocument
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING
from ftw.solr.tests.utils import get_data
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
import json
import unittest


class TestSolrDocument(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.doc = SolrDocument(json.loads(get_data('doc.json')))

    def test_string_values_are_bytes(self):
        self.assertTrue(isinstance(self.doc.path, str))
        self.assertTrue(isinstance(self.doc.allowedRolesAndUsers[0], str))

    def test_getpath(self):
        self.assertEqual(
            self.doc.getPath(), '/plone/my-folder-1/my-document.docx')

    def test_geturl(self):
        self.assertEqual(
            self.doc.getURL(),
            'http://nohost/plone/my-folder-1/my-document.docx')

    def test_get_relative_url(self):
        self.assertEqual(
            self.doc.getURL(relative=True),
            '/plone/my-folder-1/my-document.docx')

    def test_getobject(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        portal.invokeFactory('Document', 'my-document.docx')
        obj = portal['my-document.docx']
        self.doc.data[u'path'] = '/'.join(obj.getPhysicalPath())
        self.assertEqual(self.doc.getObject(), obj)

    def test_getobject_for_not_existing_object(self):
        self.assertEqual(self.doc.getObject(), None)

    def test_getobject_for_acquired_object_at_wrong_path(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        portal.invokeFactory('Document', 'my-document.docx')
        portal.invokeFactory('Folder', 'my-folder-1')
        self.assertEqual(self.doc.getObject(), None)

    def test_get_missing_field_raises(self):
        self.assertFalse(hasattr(self.doc, "foo"))
        with self.assertRaises(AttributeError):
            self.doc.foo

    def test_get_field_not_in_data_returns_none(self):
        doc = SolrDocument(json.loads(get_data('doc.json')), fields=["foo"])
        self.assertTrue(hasattr(doc, "foo"))
        self.assertIsNone(doc.foo)



# -*- coding: utf-8 -*-
from ftw.solr.browser.maintenance import ellipsified_join
import unittest


class TestEllipsifiedJoin(unittest.TestCase):

    def test_max_equal_to_item_length(self):
        self.assertEqual(
            'eins, zwo',
            ellipsified_join(['eins', 'zwo'], 2)
        )

    def test_max_smaller_than_item_length(self):
        self.assertEqual(
            'eins, ...',
            ellipsified_join(['eins', 'zwo'], 1)
        )

    def test_max_none(self):
        self.assertEqual(
            'eins, zwo, drei',
            ellipsified_join(['eins', 'zwo', 'drei'], None)
        )



# -*- coding: utf-8 -*-
from DateTime import DateTime
from ftw.solr.connection import SolrResponse
from ftw.solr.contentlisting import SolrContentListing
from ftw.solr.contentlisting import SolrContentListingObject
from ftw.solr.contentlisting import SolrDocument
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.schema import SolrSchema
from ftw.solr.testing import FTW_SOLR_INTEGRATION_TESTING
from ftw.solr.tests.utils import get_data
from mock import MagicMock
from mock import PropertyMock
from plone.app.contentlisting.interfaces import IContentListing
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from zope.interface import alsoProvides
import json
import unittest


class TestContentListing(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.resp = SolrResponse(body=get_data('search.json'), status=200)
        self.contentlisting = SolrContentListing(self.resp)

    def test_contentlisting_creation(self):
        contentlisting = IContentListing(self.resp)
        self.assertTrue(isinstance(contentlisting, SolrContentListing))

    def test_contentlisting_actual_result_count(self):
        self.assertEqual(self.contentlisting.actual_result_count, 3)

    def test_contentlisting_getitem(self):
        self.assertTrue(
            IContentListingObject.providedBy(self.contentlisting[0]))

    def test_contentlisting_iteration(self):
        self.assertTrue(IContentListingObject.providedBy(
            [item for item in self.contentlisting][0]))


class TestContentListingIntegration(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        self.manager = MagicMock(name='SolrConnectionManager')
        alsoProvides(self.manager, ISolrConnectionManager)

        conn = MagicMock(name='SolrConnection')
        conn.get = MagicMock(name='get', return_value=SolrResponse(
            body=get_data('schema.json'), status=200))
        type(self.manager).connection = PropertyMock(return_value=conn)
        type(self.manager).schema = PropertyMock(
            return_value=SolrSchema(self.manager))

        sm = self.portal.getSiteManager()
        sm.registerUtility(self.manager, ISolrConnectionManager)

        self.resp = SolrResponse(body=get_data('search.json'), status=200)
        self.contentlisting = SolrContentListing(self.resp)

    def test_contentlisting_object_documents_are_aware_of_existing_fields(self):
        doc1 = self.contentlisting[0]
        self.assertIn("path_depth", self.contentlisting.fields)
        self.assertNotIn("path_depth", doc1.doc.data)
        self.assertTrue(hasattr(doc1.doc, "path_depth"))
        self.assertIsNone(doc1.doc.path_depth)


class TestContentListingObject(unittest.TestCase):

    layer = FTW_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.obj = SolrContentListingObject(SolrDocument(
            json.loads(get_data('doc.json'))))

    def test_idublincore_title(self):
        self.assertEqual(self.obj.Title(), 'My Document')

    def test_idublincore_description(self):
        self.assertEqual(self.obj.Description(), 'Foo bar baz.')

    def test_idublincore_type(self):
        self.assertEqual(self.obj.Type(), 'Document')

    def test_idublincore_created(self):
        self.assertEqual(
            self.obj.created(), DateTime('2017/08/01 13:17:0.009000 GMT+2'))

    def test_idublincore_modified(self):
        self.assertEqual(
            self.obj.modified(), DateTime('2017/12/31 12:17:0.137000 GMT+1'))

    def test_idublincore_effective(self):
        self.assertEqual(
            self.obj.effective(), None)

    def test_idublincore_expires(self):
        self.assertEqual(
            self.obj.expires(), None)

    def test_idublincore_creationdate(self):
        self.assertEqual(
            self.obj.CreationDate(), '2017-08-01T13:17:00+02:00')

    def test_idublincore_modificationdate(self):
        self.assertEqual(
            self.obj.ModificationDate(), '2017-12-31T12:17:00+01:00')

    def test_idublincore_effectivedate(self):
        self.assertEqual(
            self.obj.EffectiveDate(), None)

    def test_idublincore_expirationdate(self):
        self.assertEqual(
            self.obj.ExpirationDate(), None)

    def test_icontentlistingobject_getid(self):
        self.assertEqual(self.obj.getId(), 'my-document.docx')

    def test_icontentlistingobject_getobject(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        portal.invokeFactory('Document', 'my-document')
        obj = portal['my-document']
        self.obj.doc.data[u'path'] = '/'.join(obj.getPhysicalPath())
        self.assertEqual(self.obj.getObject(), obj)

    def test_icontentlistingobject_getobject_for_not_existing_object(self):
        self.assertEqual(self.obj.getObject(), None)

    def test_icontentlistingobject_getdataorigin(self):
        self.assertTrue(isinstance(self.obj.getDataOrigin(), SolrDocument))

    def test_icontentlistingobject_getpath(self):
        self.assertEqual(
            self.obj.getPath(), '/plone/my-folder-1/my-document.docx')

    def test_icontentlistingobject_geturl(self):
        self.assertEqual(
            self.obj.getURL(),
            'http://nohost/plone/my-folder-1/my-document.docx')

    def test_icontentlistingobject_uuid(self):
        self.assertEqual(
            self.obj.uuid(), '9398dad21bcd49f8a197cd50d10ea778')

    def test_icontentlistingobject_geticon(self):
        self.assertEqual(
            self.obj.getIcon(),
            '<img width="16" height="16" '
            'src="http://nohost/plone/icon_dokument_word.gif" '
            'alt="Page Word 2007 document" />')

    def test_icontentlistingobject_getsize(self):
        self.assertEqual(self.obj.getSize(), '374 KB')

    def test_icontentlistingobject_review_state(self):
        self.assertEqual(self.obj.review_state(), 'private')

    def test_icontentlistingobject_portaltype(self):
        self.assertEqual(self.obj.PortalType(), 'Document')

    def test_icontentlistingobject_croppeddescription(self):
        self.assertEqual(self.obj.CroppedDescription(), 'Foo bar baz.')

    def test_icontentlistingobject_contenttypeclass(self):
        self.assertEqual(self.obj.ContentTypeClass(), 'contenttype-document')



from DateTime import DateTime
from ftw.solr.document import SolrDocument
from ftw.solr.interfaces import ISolrConnectionManager
from plone.app.contentlisting.contentlisting import BaseContentListingObject
from plone.app.contentlisting.interfaces import IContentListing
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.layout.icons.interfaces import IContentIcon
from time import localtime
from zope.component import getUtility
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.interface import implementer


@implementer(IContentListing)
class SolrContentListing(object):

    doc_type = SolrDocument

    def __init__(self, resp):
        self.resp = resp
        manager = getUtility(ISolrConnectionManager)
        self.fields = set(manager.schema.fields.keys())

    def __getitem__(self, key):
        return self._create_content_listing_object(self.resp.docs[key])

    def __len__(self):
        return len(self.resp.docs)

    @property
    def actual_result_count(self):
        return self.resp.num_found

    def __iter__(self):
        for doc in self.resp.docs:
            yield self._create_content_listing_object(doc)

    def _create_content_listing_object(self, doc):
        doc = self.doc_type(doc, self.fields)
        self._add_snippets(doc)
        return IContentListingObject(doc)

    def _add_snippets(self, doc):
        snippets = ''
        if 'highlighting' in self.resp:
            for snippet in self.resp['highlighting'].get(doc.UID).values():
                snippets += ' '.join(snippet)
        doc['_snippets_'] = snippets


@implementer(IContentListingObject)
class SolrContentListingObject(BaseContentListingObject):

    def __init__(self, doc):
        self.doc = doc

    @property
    def snippets(self):
        return self.doc.get('_snippets_')

    def Title(self):
        return self.doc.get('Title')

    def Description(self):
        return self.doc.get('Description')

    def Type(self):
        return self.doc.portal_type

    def created(self, zone=None):
        return self._zope_datetime('created', zone=zone)

    def modified(self, zone=None):
        return self._zope_datetime('modified', zone=zone)

    def effective(self, zone=None):
        return self._zope_datetime('effective', zone=zone)

    def expires(self, zone=None):
        return self._zope_datetime('expires', zone=zone)

    def CreationDate(self, zone=None):
        return self._dcmi_date('created', zone=zone)

    def ModificationDate(self, zone=None):
        return self._dcmi_date('modified', zone=zone)

    def EffectiveDate(self, zone=None):
        return self._dcmi_date('effective', zone=zone)

    def ExpirationDate(self, zone=None):
        return self._dcmi_date('expires', zone=zone)

    def _zope_datetime(self, field, zone=None):
        dt = self.doc.get(field)
        if dt is None:
            return None
        dt = DateTime(dt)
        if zone is None:
            zone = DateTime().localZone(localtime(dt.timeTime()))
        return dt.toZone(zone)

    def _dcmi_date(self, field, zone=None):
        dt = self._zope_datetime(field, zone=zone)
        if dt is not None:
            return dt.ISO8601()

    def getId(self):
        return self.doc.get('id')

    def getObject(self, REQUEST=None, restricted=True):
        return self.doc.getObject(REQUEST=REQUEST, restricted=restricted)

    def getDataOrigin(self):
        return self.doc

    def getPath(self):
        return self.doc.get('path')

    def getURL(self, relative=False):
        return self.doc.getURL(relative=relative)

    def uuid(self):
        return self.doc.get('UID')

    def getIcon(self):
        return queryMultiAdapter((getSite(), getRequest(), self.doc),
                                 interface=IContentIcon)()

    def getSize(self):
        return self.doc.get('getObjSize')

    def review_state(self):
        return self.doc.get('review_state')

    def PortalType(self):
        return self.doc.get('portal_type')

    def CroppedDescription(self):
        return self.doc.get('Description')



# -*- coding: utf-8 -*-
from Products.CMFPlone.utils import getFSVersionTuple
from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.schema import Bool
from zope.schema import Text


PLONE51 = getFSVersionTuple() >= (5, 1)
if PLONE51:
    from Products.CMFCore.interfaces import IIndexQueueProcessor
else:
    from collective.indexing.interfaces import IIndexQueueProcessor


class IFtwSolrLayer(IDefaultBrowserLayer):
    """Marker interface that defines a browser layer."""


class ISolrIndexQueueProcessor(IIndexQueueProcessor):
    """Solr index queue processor"""


class ISolrConnectionManager(Interface):
    """Solr connection manager"""


class ISolrConnectionConfig(Interface):
    """Solr connection configuration"""


class ISolrIndexHandler(Interface):
    """Handles uploading of data in Solr"""


class ISolrSearch(Interface):
    """Solr search utility"""

    def search(request_handler=u'/select', query=u'*:*', filters=None,
               start=0, rows=1000, sort=None, **params):
        """Perform a search with the given querystring and extra parameters"""


class ISolrDocument(Interface):
    """Solr document"""


class ISolrSettings(Interface):

    enabled = Bool(
        title=u'Enable Indexing',
        description=u'Check this to enable indexing in Solr.',
        default=True,
    )

    local_query_parameters = Text(
        title=u'Local Query Parameters',
        description=u"Prefixes the query string with local parameters. Must "
                    u"begin with '{!' and end with '}'.",
        default=u'{!boost b=recip(ms(NOW,modified),3.858e-10,10,1)}',
        required=False,
    )

    simple_search_term_pattern = Text(
        title=u'Simple Search Term Pattern',
        description=u'Pattern applied to every term of a simple search query.'
                    u'{term} will be replaced with the search term.',
        default=u'Title:{term}^100 OR Title:{term}*^10 '
                u'OR SearchableText:{term}^10 OR SearchableText:{term}*',
    )

    simple_search_phrase_pattern = Text(
        title=u'Simple Search Phrase Pattern',
        description=u'Pattern used with the whole search phrase in a simple '
                    u'search query. {phrase} will be replaced with the search '
                    u'phrase. Use this to boost the whole phrase more than the'
                    u'single terms.',
        default=u'Title:"{phrase}"^500 OR Title:"{phrase}*"^200 '
                u'OR SearchableText:"{phrase}"^200 '
                u'OR SearchableText:"{phrase}*"^20',
    )

    complex_search_pattern = Text(
        title=u'Complex Search Pattern',
        description=u'Pattern for complex search queries containing boolean '
                    u'operators. {phrase} will be replaced with the search '
                    u'query.',
        default=u'Title:({phrase})^10 OR SearchableText:({phrase})',
    )



from functools import partial
from itertools import izip


MAXCHUNKSIZE = 1 << 16


def group_by_two(iterable):
    # group_by_two('ABCDEFG') --> AB CD EF
    args = [iter(iterable)] * 2
    return izip(*args)


def chunked_file_reader(open_file, chunk_size=MAXCHUNKSIZE):
    for chunk in iter(partial(open_file.read, chunk_size), ''):
        yield chunk


def http_chunked_encoder(iterable):
    for data in iterable:
        chunk_size = '%s\r\n' % hex(len(data))[2:]
        chunk_data = '%s\r\n' % data
        chunk = chunk_size + chunk_data
        yield chunk

    # last chunk
    yield '0\r\n\r\n'



from DateTime import DateTime
from datetime import datetime
from datetime import date
import math


def to_iso8601(value, multivalued=False):
    if isinstance(value, DateTime):
        v = value.toZone('UTC')
        millisecond, second = math.modf(v.second())
        value = u'%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
            v.year(), v.month(), v.day(), v.hour(), v.minute(),
            int(second), millisecond * 1000
        )
    elif isinstance(value, datetime):
        if value.tzinfo is not None:
            # Convert to timezone naive in UTC
            value = (value - value.utcoffset()).replace(tzinfo=None)

        value = u'%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
            value.year, value.month, value.day, value.hour, value.minute,
            value.second, value.microsecond / 1000
        )
    elif isinstance(value, date):
        value = u'%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
            value.year, value.month, value.day, 0, 0, 0, 0
        )
    else:
        value = None
    return value


def to_unicode(value, multivalued=False):
    if isinstance(value, tuple):
        value = list(value)
    if not isinstance(value, list):
        value = [value]
    for i, v in enumerate(value):
        if isinstance(v, str):
            v = v.decode('utf8')
        value[i] = v
    if not multivalued:
        if len(value) > 1:
            return u' '.join(value)
        else:
            return value[0]
    return value


CONVERTERS = {
    'solr.DatePointField': to_iso8601,
    'solr.StrField': to_unicode,
}



from datetime import date
from DateTime import DateTime
from datetime import datetime
from ftw.solr.browser.maintenance import checkpoint_iterator
from ftw.solr.browser.maintenance import timer
from ftw.solr.converters import to_iso8601
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrIndexHandler
from ftw.solr.interfaces import ISolrSettings
from ftw.upgrade import UpgradeStep
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from time import clock
from zope.component import getMultiAdapter
from zope.component import queryUtility
import logging


def _pre_datetime_format_fix_to_iso8601(value, multivalued=False):
    if isinstance(value, DateTime):
        v = value.toZone('UTC')
        value = u'%04d-%02d-%02dT%02d:%02d:%06.3fZ' % (
            v.year(), v.month(), v.day(), v.hour(), v.minute(), v.second()
        )
    elif isinstance(value, datetime):
        if value.tzinfo is not None:
            # Convert to timezone naive in UTC
            value = (value - value.utcoffset()).replace(tzinfo=None)

        value = u'%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
            value.year, value.month, value.day, value.hour, value.minute,
            value.second, value.microsecond / 1000
        )
    elif isinstance(value, date):
        value = u'%04d-%02d-%02dT%02d:%02d:%02d.%03dZ' % (
            value.year, value.month, value.day, 0, 0, 0, 0
        )
    else:
        value = None
    return value


def _solr_date(date):
    """Solr date representation. Fractional seconds are stripped if 0."""
    value = to_iso8601(date)
    if value.endswith('.000Z'):
        value = value[:-5] + 'Z'
    return value


def _pre_datetime_format_fix_solr_date(date):
    """Solr date representation. Fractional seconds are stripped if 0."""
    value = _pre_datetime_format_fix_to_iso8601(date)
    if value.endswith('.000Z'):
        value = value[:-5] + 'Z'
    return value


commit_interval = 100


class UpdateModifiedIndex(UpgradeStep):
    """Update modified index.
    """

    deferrable = True

    def __call__(self):
        self.logger = logging.getLogger('ftw.upgrade')

        if not self.is_enabled():
            self.log("Skipping: Solr is not enabled.")
            return
        self.sync()

    def is_enabled(self):
        registry = queryUtility(IRegistry)
        settings = registry.forInterface(ISolrSettings)
        return settings.enabled

    def get_objects_that_need_syncing_of_modified(self):
        """Find objects that are not in sync only due to change in DateTime
        rounding"""

        catalog = getToolByName(self.portal, 'portal_catalog')
        items = catalog.unrestrictedSearchResults()
        catalog_modified = set(
            [(item.UID, _solr_date(item.modified)) for item in items])
        pre_datetime_format_fix_catalog_modified = set(
            [(item.UID, _pre_datetime_format_fix_solr_date(item.modified)) for item in items])

        conn = self.manager.connection
        resp = conn.search({
            u'query': u'*:*',
            u'limit': 10000000,
            u'params': {u'fl': ['UID', 'modified']},
        })
        solr_modified = set(
            [(doc['UID'], doc.get('modified', u'2000-01-01T00:00:00.000Z'))
             for doc in resp.docs])

        not_in_sync = set(
            (item[0] for item in catalog_modified - solr_modified))
        pre_datetime_format_fix_not_in_sync = set(
            (item[0] for item in pre_datetime_format_fix_catalog_modified - solr_modified))

        # We only want to reindex modified for objects that would
        # be in sync if we hadn't modified the rounding in solr_date
        to_correct = not_in_sync - pre_datetime_format_fix_not_in_sync
        return to_correct

    def sync(self, commit_interval=100, idxs=None):
        """Sync Solr with portal catalog"""
        if not self.is_enabled():
            return 'Solr indexing is disabled.'

        to_correct = self.get_objects_that_need_syncing_of_modified()

        # Reindex modified for Solr items that should be in sync
        processed = 0
        real = timer()
        lap = timer()
        cpu = timer(clock)

        def commit():
            conn = self.manager.connection
            conn.commit(soft_commit=False, extract_after_commit=False)
            self.log(
                'Intermediate commit (%d items processed, last batch in %s)',
                processed, next(lap))

        cpi = checkpoint_iterator(commit, interval=commit_interval)
        self.log('Syncing modified to solr for {} objects'.format(len(to_correct)))
        catalog = getToolByName(self.portal, 'portal_catalog')
        for uid in to_correct:
            catalog_item = catalog.unrestrictedSearchResults(UID=uid)[0]

            handler = getMultiAdapter((catalog_item, self.manager), ISolrIndexHandler)
            handler.add(['modified'])
            processed += 1
            next(cpi)

        commit()

        conn = self.manager.connection
        conn.commit(soft_commit=False)

        self.log('Solr modified index synced.')
        self.log(
            'Processed %d items in %s (%s cpu time).',
            processed, next(real), next(cpu))

    @property
    def manager(self):
        return queryUtility(ISolrConnectionManager)

    def log(self, msg, *args):
        self.logger.info(msg, *args)



# -*- coding: utf-8 -*-
from DateTime import DateTime
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrSettings
from ftw.solr.query import escape
from ftw.solr.query import is_simple_search
from ftw.solr.query import make_filters
from ftw.solr.query import make_path_filter
from ftw.solr.query import make_query
from ftw.solr.query import split_simple_search
from mock import Mock
from plone.registry import Registry
from plone.registry.fieldfactory import persistentFieldAdapter
from plone.registry.interfaces import IRegistry
from plone.testing import zca
from zope.component import provideAdapter
from zope.component import provideUtility
import unittest


class TestQueryHelpers(unittest.TestCase):

    def test_escape(self):
        self.assertEqual(escape('foo!'), 'foo\\!')
        self.assertEqual(escape('1:1'), '1\\:1')
        self.assertEqual(escape('/path/to/a/file'), '\\/path\\/to\\/a\\/file')

    def test_is_simple_search(self):
        self.assertTrue(is_simple_search('foo'))
        self.assertTrue(is_simple_search('foo bar'))
        self.assertTrue(is_simple_search('"foo bar"'))
        self.assertTrue(is_simple_search('foo "bar foobar" baz'))
        self.assertTrue(is_simple_search('"foo AND bar"'))
        self.assertTrue(is_simple_search('foo "AND" bar'))
        self.assertTrue(is_simple_search('foo"'))

        self.assertFalse(is_simple_search('foo AND bar'))
        self.assertFalse(is_simple_search('foo OR bar'))
        self.assertFalse(is_simple_search('"foo bar" OR baz'))

    def test_split_simple_search(self):
        self.assertEqual(split_simple_search('foo'), ['foo'])
        self.assertEqual(split_simple_search('foo bar'), ['foo', 'bar'])
        self.assertEqual(split_simple_search('"foo bar"'), ['"foo bar"'])
        self.assertEqual(
            split_simple_search('foo "bar foobar" baz'),
            ['foo', '"bar foobar"', 'baz'])


class TestMakeQuery(unittest.TestCase):

    layer = zca.UNIT_TESTING

    def setUp(self):
        provideAdapter(persistentFieldAdapter)
        registry = Registry()
        registry.registerInterface(ISolrSettings)
        provideUtility(registry, IRegistry)
        self.settings = registry.forInterface(ISolrSettings)
        self.settings.simple_search_term_pattern = (
            u'Title:{term}^10 OR SearchableText:{term} OR '
            u'SearchableText:{term}*')
        self.settings.simple_search_phrase_pattern = (
            u'Title:"{phrase}"^20 OR SearchableText:"{phrase}"^5 OR '
            u'SearchableText:"{phrase}*"^2')
        self.settings.complex_search_pattern = (
            u'Title:({phrase})^10 OR SearchableText:({phrase})')
        self.settings.local_query_parameters = u''

    def test_single_search_word(self):
        self.assertEqual(
            make_query('foo'),
            u'Title:foo^10 OR SearchableText:foo OR SearchableText:foo*')

    def test_multiple_search_words(self):
        self.assertEqual(
            make_query('foo bar baz'),
            u'(Title:"foo bar baz"^20 OR SearchableText:"foo bar baz"^5 OR '
            u'SearchableText:"foo bar baz*"^2) OR ('
            u'(Title:foo^10 OR SearchableText:foo OR SearchableText:foo*) AND '
            u'(Title:bar^10 OR SearchableText:bar OR SearchableText:bar*) AND '
            u'(Title:baz^10 OR SearchableText:baz OR SearchableText:baz*))')

    def test_search_words_with_binary_operator(self):
        self.assertEqual(
            make_query('foo AND bar'),
            u'Title:(foo AND bar)^10 OR SearchableText:(foo AND bar)')

    def test_search_word_with_special_chars(self):
        self.assertEqual(
            make_query('C++'),
            u'(Title:"C\\+\\+"^20 OR SearchableText:"C\\+\\+"^5 OR '
            u'SearchableText:"C\\+\\+*"^2) OR '
            u'(Title:C\\+\\+^10 OR SearchableText:C\\+\\+ OR '
            u'SearchableText:C\\+\\+*)')

    def test_query_with_local_parameters(self):
        self.settings.local_query_parameters = (
            u'{!boost b=recip(ms(NOW,modified),3.858e-10,10,1)}')
        self.assertEqual(
            make_query('foo'),
            u'{!boost b=recip(ms(NOW,modified),3.858e-10,10,1)}'
            u'Title:foo^10 OR SearchableText:foo OR SearchableText:foo*')

    def test_long_phrase_gets_truncated_to_10_terms(self):
        self.settings.simple_search_term_pattern = u'Title:{term}'
        self.settings.simple_search_phrase_pattern = u'Title:"{phrase}"'
        self.assertEqual(
            make_query(
                'one two three four five six seven eight nine ten eleven'),
            u'(Title:"one two three four five six seven eight nine ten '
            u'eleven") OR ((Title:one) AND (Title:two) AND (Title:three) AND '
            u'(Title:four) AND (Title:five) AND (Title:six) AND (Title:seven) '
            u'AND (Title:eight) AND (Title:nine) AND (Title:ten))')

    def test_query_string_is_unicode(self):
        self.assertTrue(isinstance(make_query('über'), unicode))
        self.assertTrue(isinstance(make_query(u'über'), unicode))


class TestMakeFilters(unittest.TestCase):

    layer = zca.UNIT_TESTING

    def setUp(self):
        manager = Mock()
        manager.schema.fields = {
            'UID': None,
            'Title': None,
            'is_folderish': None,
            'modified': None,
            'path': None,
        }
        provideUtility(manager, ISolrConnectionManager)

    def test_make_filters_from_string(self):
        self.assertEqual(
            make_filters(UID='85bed8c49f6d4f8b841693c6a7c6cff1'),
            [u'UID:85bed8c49f6d4f8b841693c6a7c6cff1'],
        )

    def test_make_filters_from_list(self):
        self.assertEqual(
            make_filters(UID=[
                '85bed8c49f6d4f8b841693c6a7c6cff1',
                '85a29144758f494c88df19182f749ed6',
            ]),
            ['UID:(85bed8c49f6d4f8b841693c6a7c6cff1 OR '
             '85a29144758f494c88df19182f749ed6)'],
        )

    def test_make_filters_from_boolean(self):
        self.assertEqual(
            make_filters(is_folderish=True), ['is_folderish:true'])
        self.assertEqual(
            make_filters(is_folderish=False), ['is_folderish:false'])

    def test_make_filters_from_date_range(self):
        self.assertEqual(
            make_filters(modified={
                'query': DateTime('2020-05-27T10:00:00'),
                'range': 'min',
            }),
            ['modified:[2020\\-05\\-27T10\\:00\\:00.000Z TO *]'],
        )
        self.assertEqual(
            make_filters(modified={
                'query': DateTime('2019-05-27T10:00:00'),
                'range': 'max',
            }),
            ['modified:[* TO 2019\\-05\\-27T10\\:00\\:00.000Z]'],
        )
        self.assertEqual(
            make_filters(modified={
                'query': [
                    DateTime('2020-05-27T10:00:00'),
                    DateTime('2019-05-27T10:00:00'),
                ],
                'range': 'minmax',
            }),
            ['modified:[2019\\-05\\-27T10\\:00\\:00.000Z TO '
             '2020\\-05\\-27T10\\:00\\:00.000Z]'],
        )

    def test_make_filters_from_path(self):
        self.assertEqual(
            make_filters(path='/plone/folder'),
            ['path:\\/plone\\/folder'],
        )
        self.assertEqual(
            make_filters(path={'query': '/plone/folder', 'depth': 1}),
            ['path_parent:\\/plone\\/folder', 'path_depth:3'],
        )

    def test_make_filters_ignores_fields_not_in_schema(self):
        self.assertEqual(make_filters(unknown_field='foo bar'), [])

    def test_make_filters_with_utf8_chars(self):
        self.assertEqual(make_filters(Title='schön'), [u'Title:schön'])
        self.assertEqual(
            make_filters(Title=['schön', 'schöner']),
            [u'Title:(schön OR schöner)']
        )
        self.assertEqual(
            make_filters(Title={
                'query': ['schön', 'schöner'],
                'operator': 'AND',
            }),
            [u'Title:(schön AND schöner)'],
        )

    def test_make_path_filter(self):
        self.assertEqual(
            make_path_filter('/plone/folder'),
            ['path:\\/plone\\/folder'],
        )
        self.assertEqual(
            make_path_filter('/plone/folder', depth=-1),
            ['path_parent:\\/plone\\/folder'],
        )
        self.assertEqual(
            make_path_filter('/plone/folder', depth=1),
            ['path_parent:\\/plone\\/folder', 'path_depth:3'],
        )
        self.assertEqual(
            make_path_filter('/plone/folder', depth=2),
            ['path_parent:\\/plone\\/folder', 'path_depth:[3 TO 4]'],
        )
        self.assertEqual(
            make_path_filter('/plone/folder', depth=2, include_self=True),
            ['path_parent:\\/plone\\/folder', 'path_depth:[2 TO 4]'],
        )





class SolrSchema(object):

    def __init__(self, manager):
        self.manager = manager
        self.retrieve()

    def retrieve(self):
        """Retrieve the Solr schema"""
        conn = self.manager.connection
        if conn is None:
            schema = {}
        else:
            resp = conn.get('/schema')
            schema = resp.get(u'schema', {})
        self.unique_key = schema.get(u'uniqueKey')
        self.name = schema.get(u'name')
        self.version = schema.get(u'version')
        self.fields = {}
        for field in schema.get(u'fields', []):
            self.fields[field[u'name']] = field
        self.copy_fields = {}
        for field in schema.get(u'copyFields', []):
            self.copy_fields[field[u'dest']] = field
        self.dynamic_fields = {}
        for field in schema.get(u'dynamicFields', []):
            self.dynamic_fields[field[u'name']] = field
        self.field_types = {}
        for ftype in schema.get(u'fieldTypes', []):
            self.field_types[ftype[u'name']] = ftype

    def __nonzero__(self):
        if self.unique_key is not None:
            return True
        return False

    def field_class(self, name):
        return self.field_types.get(
                self.fields.get(name, {}).get(u'type'), {}).get(u'class')



# -*- coding: utf-8 -*-
"""Installer for the ftw.solr package."""

from setuptools import find_packages
from setuptools import setup
import os


long_description = '\n\n'.join([
    open('README.rst').read(),
    open(os.path.join("docs", "HISTORY.txt")).read(),
])

tests_require = [
    'plone.app.testing',
    'plone.app.dexterity',
    'plone.testing',
    'plone.api',
    'plone.app.contenttypes',
    'pytz',
    'mock',
    'ftw.testing',
]

setup(
    name='ftw.solr',
    version='2.9.5',
    description="Solr integration for Plone",
    long_description=long_description,
    # Get more from https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 4.3",
        "Framework :: Plone :: 5.0",
        "Framework :: Plone :: 5.1",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    ],
    keywords='Plone Solr',
    author='Thomas Buchberger',
    author_email='t.buchberger@4teamwork.ch',
    url='https://pypi.python.org/pypi/ftw.solr',
    license='GPL version 2',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['ftw'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        # -*- Extra requirements: -*-
        'setuptools',
        'plone.app.contentlisting',
        'plone.app.layout',
        'plone.namedfile[blobs]',
        'ftw.upgrade',
    ],
    extras_require=dict(
        test=tests_require,
        tests=tests_require,
        cindexing=['collective.indexing'],
    ),
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    [zopectl.command]
    solr = ftw.solr.commands:solr
    """,
)



from AccessControl.SecurityManagement import getSecurityManager
from ftw.solr.interfaces import ISolrConnectionManager
from ftw.solr.interfaces import ISolrSearch
from ftw.solr.query import escape
from Products.CMFPlone.utils import base_hasattr
from zope.component import queryUtility
from zope.interface import implementer


@implementer(ISolrSearch)
class SolrSearch(object):
    """A search utility for Solr """

    def __init__(self):
        self._manager = None

    @property
    def manager(self):
        if self._manager is None:
            self._manager = queryUtility(ISolrConnectionManager)
        return self._manager

    def _search(self, request_handler=u'/select', query=u'*:*', filters=None,
                start=0, rows=1000, sort=None, unrestricted=False, **params):
        conn = self.manager.connection
        params = {u'params': params}
        params[u'query'] = query
        params[u'offset'] = start
        params[u'limit'] = rows
        if sort is not None:
            params[u'sort'] = sort
        if filters is None:
            filters = []
        if not isinstance(filters, list):
            filters = [filters]
        if not unrestricted:
            filters.insert(0, self.security_filter())
        params[u'filter'] = filters
        return conn.search(params, request_handler=request_handler)

    def search(self, request_handler=u'/select', query=u'*:*', filters=None,
               start=0, rows=1000, sort=None, **params):
        return self._search(
            request_handler=request_handler, query=query, filters=filters,
            start=start, rows=rows, sort=sort, unrestricted=False, **params)

    def unrestricted_search(self, request_handler=u'/select', query=u'*:*',
                            filters=None, start=0, rows=1000, sort=None, **params):
        return self._search(
            request_handler=request_handler, query=query, filters=filters,
            start=start, rows=rows, sort=sort, unrestricted=True, **params)

    def security_filter(self):
        user = getSecurityManager().getUser()
        roles = user.getRoles()
        if 'Anonymous' in roles:
            return u'allowedRolesAndUsers:Anonymous'
        roles = list(roles)
        roles.append('Anonymous')
        if base_hasattr(user, 'getGroups'):
            groups = [u'user:%s' % x for x in user.getGroups()]
            if groups:
                roles = roles + groups
        roles.append(u'user:%s' % user.getId())
        # Roles with spaces need to be quoted
        roles = [u'"%s"' % escape(r) if ' ' in r else escape(r) for r in roles]
        return u'allowedRolesAndUsers:(%s)' % u' OR '.join(roles)
